"""a simple container for a list of transformers

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""

from primrose.base.transformer import AbstractTransformer

class TransformerSequence():
    """A container for list of transformers"""

    def __init__(self, sequence=[]):
        """init a transformer sequence

        Args:
            sequence (list of transformers): list of transformers

        Returns:
            nothing. Transformers in sequence are added to the internal list

        """
        self.sequence = []
        
        for s in sequence:
            self.add(s)

    def add(self, transformer):
        """add a transformer to the sequence

        Returns:
            nothing. Side effect is to add transformer to the list

        Raises:
            Exception if transformer is not a transformer

        """
        if not isinstance(transformer, AbstractTransformer):
            raise Exception("Transformer needs to extend AbstractTransformer")

        self.sequence.append(transformer)

    def transformers(self):
        """generate: yield each transformer

        Yields:
            yields a transformer from the list

        """
        for transformer in self.sequence:
            yield transformer



"""Module to traverse DAG by running depth first algorithm

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import networkx as nx
from primrose.dag.dag_traverser import DagTraverser

class DepthFirstTraverser(DagTraverser):
    """a traverser that traverses in depth first manner"""

    def run_section_by_section(self):
        """Do we want go through section by section?

        Note:
            suppose we had 5 sections and we asked to run just 2: section1 and section2.
            Is it important that we run section1 and then section 2 (if so, return True)
            or just that we run all the nodes from section1 and section2 together in some
            desired order (return False)

        Returns:
            result (Boolean)

        """
        return False

    def traversal_list(self):
        """get list of nodes to traverse doing a depth first traversal

        Returns:
            sequence (list): list of node name in the order that they will be run

        """
        G = self.configuration.dag.G2
        sequence = list(nx.dfs_postorder_nodes(G))
        sequence.reverse()
        return sequence




import pytest
from primrose.readers.gcs_dill_reader import GcsDillReader
from primrose.data_object import DataObject
from primrose.readers.csv_reader import CsvReader
from primrose.configuration.configuration import Configuration
import dill
import os

def test_necessary_config():
    assert len(GcsDillReader.necessary_config({})) == 2

def test_run(monkeypatch):
    # returns 2 objects from dill reader
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDillReader",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDillReader(configuration, "myreader")

    data_object = DataObject(configuration)

    def fake_blobs():
        with open("test/dill_reader_blob1.pkl", "wb") as dill_file:
            dill.dump("some_data", dill_file)
        with open("test/dill_reader_blob2.pkl", "wb") as dill_file:
            dill.dump("some_other_data", dill_file)
        dat1 = open("test/dill_reader_blob1.pkl", "rb").read()
        dat2 = open("test/dill_reader_blob2.pkl", "rb").read()
        return [dat1, dat2]

    monkeypatch.setattr(reader,'download_blobs_as_strings',fake_blobs)

    reader_object, terminate = reader.run(data_object)

    assert not terminate

    assert "myreader" in reader_object.data_dict

    dat = reader_object.data_dict['myreader']

    assert "reader_data" in dat
    datlist = dat['reader_data']

    assert len(datlist) == 2

    assert "some_data" in datlist
    assert "some_other_data" in datlist

    files = ["test/dill_reader_blob1.pkl", "test/dill_reader_blob2.pkl"]
    for f in files:
        if os.path.exists(f):
            os.remove(f)

def test_run2(monkeypatch):
    # returns 1 objects from dill reader
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDillReader",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDillReader(configuration, "myreader")

    data_object = DataObject(configuration)

    def fake_blobs():
        with open("test/dill_reader_blob1.pkl", "wb") as dill_file:
            dill.dump("some_data", dill_file)
        dat1 = open("test/dill_reader_blob1.pkl", "rb").read()
        return [dat1]

    monkeypatch.setattr(reader,'download_blobs_as_strings',fake_blobs)

    reader_object, terminate = reader.run(data_object)

    assert not terminate

    assert "myreader" in reader_object.data_dict

    dat = reader_object.data_dict['myreader']

    assert "reader_data" in dat
    data = dat['reader_data']

    assert "some_data" == data

    files = ["test/dill_reader_blob1.pkl"]
    for f in files:
        if os.path.exists(f):
            os.remove(f)




"""Module with abstract pipeline class to specify interface needed for future pipelines

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""

from abc import ABC, abstractmethod
from enum import Enum
from primrose.base.node import AbstractNode
import logging
from primrose.base.transformer_sequence import TransformerSequence


class PipelineModeType(Enum):
    """Mode when performing the pipeline

    Note:
        FIT = fit data to transformer object only
        TRANSFORM = transform data only from (previously) fit transformers in a pipeline
        FIT_TRANSFORM = fit data then transform data in a pipeline

    """
    FIT = "FIT"
    FIT_TRANSFORM = "FIT_TRANSFORM"
    TRANSFORM = "TRANSFORM"

    @staticmethod
    def names():
        """list of all the names in the enum

        Returns:
            list of Enum's names

        """
        return list(map(lambda t: t.name, PipelineModeType))

    @staticmethod
    def values():
        """list of the enum's values

        Returns:
            list of (value, PipelineModeType) pairs

        """
        return list(map(lambda t: t.value, PipelineModeType))


class AbstractPipeline(AbstractNode):
    """Pipeline class should have a defined pipeline that it executes and the ability to transform raw data"""

    def __init__(self, configuration, instance_name):
        """Clean/transform/filter data in memory after de-serializing from a reader object

        Args:
            configuration: configuration class specified in src/configuration
            instance_name: name used to find this specific instance's configuration

        """
        super(AbstractPipeline, self).__init__(configuration, instance_name)
        self.transformer_sequence = None
        self.data = None

    def check_for_upstream_transformers(self, data_object):
        """Examine the upstream data_object for any TransformerSequence

        if not found, then initialize a new TransformerSequence

        Args:
            data_object (DataObject): istance of DataObject

        Returns:
            TransformerSequence

        """
        # check for a valid pipeline object inside the data object
        source_dict = data_object.get_upstream_data(self.instance_name, pop_data=False)

        # look for upstream transformer objects
        for source in source_dict:

            # NOTE: some readers in primrose have objects nested TransformerSequences one level deeper than others
            # we will allow for this scenario by checking both the first and second level keys for a TransformerSequence
            if isinstance(source_dict[source], TransformerSequence):
                logging.info('Upstream TransformerSequence found, initializing pipeline...')
                return source_dict[source]
            elif isinstance(source_dict[source], dict):
                for subkey in source_dict[source]:
                    if isinstance(source_dict[source][subkey], TransformerSequence):
                        logging.info('Upstream TransformerSequence found, initializing pipeline...')
                        return source_dict[source][subkey]

        logging.info('No upstream TransformerSequence found. Creating new TransformerSequence...')

        return self.init_pipeline()

    def run(self, data_object):
        """Run pipeline on the data object

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        is_training = self.node_config.get('is_training', None)

        if is_training is None:
            logging.info('"is_training" key not found in the node_config, assuming production data')
            is_training = False
        else:
            is_training = str(is_training).lower() == "true"

        self.transformer_sequence = self.check_for_upstream_transformers(data_object)

        if is_training:
            data_object = self.fit_transform(data_object)
        else:
            data_object = self.transform(data_object)

        terminate = False
        return data_object, terminate

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        After adding this list, validation automatically occurs before instantiation in the pipeline factory.

        Returns:
            set of keys necessary to run implementation

        """
        return set(['is_training']) # pragma: no cover

    def fit_transform(self, data_object):
        """Clean/transform or filter data using a pipeline of functions

        The method should also cache results and report on sizing for debugging. fit_transform must store the
        information necessary for data transformations on test data, so any encodings or model-based imputations must be
        cached in this method, to be called when the transform method is used.

        Args:
            data_object (DataObject): DataObject instance

        Returns:
            data_object (DataObject): DataObject instance

        """
        return self.transform(data_object)

    @abstractmethod
    def transform(self, data_object):
        """Clean/transform or filter data using a pipeline of functions and any cached objects from fit_transform

        The method should cache post-transform results and report on sizing for debugging. tranform uses the cached
        objects scored in the fit_tranform call to transform data. It's likely to be used for live predictions or
        test (hold-out) data.

        Args:
            data_object (DataObject): DataObject instance

        Returns:
            data_object (DataObject)

        """
        return data_object # pragma: no cover

    def init_pipeline(self):
        """Initialize the pipeline if no pipeline object is found in the upstream data objects

        Returns:
            TransformerSequence

        """
        return TransformerSequence()

    def execute_pipeline(self, input_, mode):
        """Run a TransformerSequence of functions with chained input and output data

        Args:
            input_ (object): input data (usually a pandas dataframe)
            mode: enum object for fit, transform, or fit_transform

        Returns:
            transformed data (usually a pandas dataframe) after running through all functions in the pipeline

        """
        if not self.transformer_sequence:
            raise Exception("run() must be called to extract/create a TransformerSequence")

        if mode not in PipelineModeType.names() and mode not in PipelineModeType:
            raise Exception('mode must be of type PipelineModeType Enum object.')

        for transformer in self.transformer_sequence.transformers():

            if mode == PipelineModeType.FIT:

                transformer.fit(input_)

            elif mode == PipelineModeType.TRANSFORM:

                input_ = transformer.transform(input_)

            elif mode == PipelineModeType.FIT_TRANSFORM:

                input_ = transformer.fit_transform(input_)

        return input_



"""Module with AbstractNode implementation, Example model for primrose user modification

Author(s):
    Mike Skarlinski (michael.skarlinski@ww.com)

"""
from primrose.base.model import AbstractModel


class AwesomeModel(AbstractModel):
    """(EXAMPLE MODEL) Print what is going on for each abstract method implementation

    Notes:
        see AbstractModel class to understand how this model is run in each mode!

    """

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the AwesomeModel object

        Put your required configuration keys here, here we pass just the base requirement from the abstract class

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            filename: name of the file

        Returns:
            set of necessary keys for the AwesomeModel object

        """
        return AbstractModel.necessary_config(node_config)

    def train_model(self, data_object):
        """Code to train your model and return a data_object after adding any training or model info"""

        # Example model training code for building your own model
        # ----------------------------------
        print('I am training my model.')

        # Example showing how to get upstream data
        # get some training data if it exists
        try:
            training_data = data_object.get_upstream_data(self.instance_name)

        except:
            print('No upstream data exists')
        # ----------------------------------

        return data_object

    def eval_model(self, data_object):
        """Code to evaluate your models performance"""

        print('My model is doing pretty well.')

        return data_object

    def predict(self, data_object):
        """Make predictions using your model"""

        print('I am predicting!')

        return data_object




from primrose.transformers.sklearn_preprocessing_transformer import SklearnPreprocessingTransformer
from sklearn.preprocessing import StandardScaler
import pytest
import pandas as pd
import math

def test_fit():
    scaler = StandardScaler(with_mean=True, with_std=True)
    t = SklearnPreprocessingTransformer(scaler, columns=['x','y'])

    df = pd.DataFrame({'x':[1,2,3,4,5], 'y': [5,20,30,40,90]})
    t.fit(df)

    transformed_data = t.transform(df)

    assert list(t.preprocessor.mean_) == [3.0, 37.0]
    assert list(t.preprocessor.var_) == [2., 836.]
    assert math.isclose(transformed_data.x.mean(), 0.0, abs_tol = 0.0001)
    #note setting this default degrees of freedom=0 is important as it won't match numpy's std for same data otherwise
    assert math.isclose(transformed_data.x.std(ddof=0), 1.0, abs_tol = 0.0001)
    assert math.isclose(transformed_data.y.mean(), 0.0, abs_tol = 0.0001)
    assert math.isclose(transformed_data.y.std(ddof=0), 1.0, abs_tol = 0.0001)



import pytest
import pandas as pd
from primrose.configuration.configuration import Configuration
from primrose.models.sklearn_classifier_model import SklearnClassifierModel
from primrose.data_object import DataObject, DataObjectResponseType
from primrose.readers.csv_reader import CsvReader
from primrose.pipelines.encode_train_test_split import EncodeTrainTestSplit


@pytest.fixture()
def config():
    return Configuration("test/hello_world_tennis.json")


@pytest.fixture()
def data_obj(config):
    df = pd.read_csv("test/tennis.csv")
    data_object = DataObject(config)
    reader = CsvReader(config, 'read_data')
    data_object.add(reader, df)
    encoder = EncodeTrainTestSplit(config, 'encode_and_split')
    data_object, terminate = encoder.run(data_object)
    return data_object


@pytest.fixture
def model(config):
    return SklearnClassifierModel(config, "decision_tree_model")


def test_necessary_config():
    assert SklearnClassifierModel.necessary_config({}) == set(["model_parameters", "mode", "sklearn_classifier_name",
                                                             "grid_search_scoring", "cv_folds"])


def test_get_data(model, data_obj):
    x_train, y_train, x_test, y_test = model._get_data(data_obj)
    assert x_train.shape[0] == y_train.shape[0]
    assert x_test.shape[0] == y_test.shape[0]  # we removed the ID column by not specifying it in X


def test_train_model(model, data_obj):
    model.train_model(data_obj)
    assert model.model is not None
    x_train, y_train, x_test, y_test = model._get_data(data_obj)
    predictions = model.model.predict(x_test)
    assert list(predictions) == [1, 0, 0, 1, 1]


def test_eval(model, data_obj):
    model.train_model(data_obj)
    model.eval_model(data_obj)
    print(model.scores)
    assert round(model.scores['recall'], 2) == 0.67
    assert round(model.scores['precision'], 2) == 0.67
    assert model.scores['predicted_class_fraction'] == 0.6
    assert model.scores['positive_class_fraction'] == 0.6


def test_predict(model, data_obj):
    model.train_model(data_obj)
    data_object = model.predict(data_obj)
    predicted = data_object.get('decision_tree_model', rtype=DataObjectResponseType.KEY_VALUE.value)

    assert predicted['predictions'].shape[0] == 5
    assert list(predicted['predictions'].predictions) == ['yes', 'no', 'no', 'yes', 'yes']



import pytest
from primrose.node_factory import NodeFactory
from primrose.base.postprocess import AbstractPostprocess
from primrose.base.node import AbstractNode
from primrose.configuration.configuration import Configuration
from primrose.data_object import DataObject, DataObjectResponseType

def test_init():
    class TestPostprocess(AbstractPostprocess):
#        def __init__(self, configuration, instance_name):
#            super(TestPostprocess, self).__init__(configuration, instance_name)
        @staticmethod
        def necessary_config(node_config):
            return set(['key1'])
        def run(self, data_object):
            data_object.add(self, "some data")
            return data_object, False
        #def process(self, data):
        #    return "some data"
    NodeFactory().register("TestPostprocess", TestPostprocess)

    class TestModel(AbstractNode):
        def __init__(self, configuration, instance_name):
            pass
        @staticmethod
        def necessary_config(node_config):
            return set(['key1'])
        def run(self, data_object):
            return data_object, False
    NodeFactory().register("TestModel", TestModel)

    config = {
        "implementation_config": {
            "model_config": {
                "modelname": {
                    "class": "TestModel",
                    "key1":"val1",
                    "destinations": ["nodename"]
                }
            },
            "postprocess_config": {
                "nodename": {
                    "class": "TestPostprocess",
                    "key1":"val1",
                    "key2":"val2",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    tp = TestPostprocess(configuration, 'nodename')
    node_config = {
                    "class": "TestPostprocess",
                    "key1":"val1",
                    "key2":"val2",
                    "destinations": []
                }
    assert set(['key1']) == tp.necessary_config(node_config)

    #assert tp.process(None) == "some data"

    data_object, terminate = tp.run(data_object)

    assert not terminate
    
    assert data_object.get("nodename", rtype=DataObjectResponseType.VALUE.value) == "some data"

#    assert tp.upstream_model_keys() == ["modelname"]

#    assert tp.single_upstream_model_key() == "modelname"



"""Module with AbstractNode implementation, able to read from local and gcs dill and pickle files

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)
    Brian Graham (brian.graham@weightwatchers.com)

"""

import dill
import logging
import pickle
from google.cloud import storage
from primrose.base.reader import AbstractReader


class Deserializer(AbstractReader):
    """Read a local file and de-serialize it into memory."""
    DATA_KEY = 'reader_data'
    SUPPORTED_DESERIALIZERS = {'dill': dill, 'pickle': pickle}

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the Deserializer object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            filename (str): local filename to be de-serialized
            deserializer (str): 'dill' or 'pickle'

        Returns:
            set of necessary keys for the Deserializer object

        """
        return set(['filename','deserializer'])

    def run(self, data_object):
        """Read dill object(s) from local filesystem

        Args:
            data_object: DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        logging.info('Reading {} from local filesystem'.format(self.node_config['filename']))

        if self.node_config['deserializer'] not in Deserializer.SUPPORTED_DESERIALIZERS.keys():
            logging.warning(f"{self.node_config['deserializer']} deserializer not supported.")
            logging.warning(f"The following deserializers are supported {Deserializer.SUPPORTED_DESERIALIZERS.keys()}.")
            raise Exception(f"Unsupported Deserializer: {self.node_config['deserializer']}")

        deserializer = Deserializer.SUPPORTED_DESERIALIZERS[self.node_config['deserializer']]

        object = deserializer.load(open(self.node_config['filename'], 'rb'))

        data_object.add(self, object, key=Deserializer.DATA_KEY)

        terminate = False

        return data_object, terminate

class GcsDeserializer(AbstractReader):
    """Read a file from GCS and de-serialize it into memory."""

    DATA_KEY = 'reader_data'
    SUPPORTED_DESERIALIZERS = {'dill': dill, 'pickle': pickle}

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the GcsDeserializer object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            bucket_name: name of the GCS bucket
            blob_name: name of the blob
            deserializer (str): 'dill' or 'pickle'

        Returns:
            set of necessary keys for the GcsDeserializer object

        """
        return set(['bucket_name', 'blob_name','deserializer'])

    def download_blobs_as_strings(self):
        """Downloads a blob from the bucket contining the user specified blob_name

        Returns:
            list of strings

        """

        if 'gcs_project' in self.node_config:
            storage_client = storage.Client(project=self.node_config['gcs_project'])
        else:
            storage_client = storage.Client()

        bucket = storage_client.get_bucket(self.node_config['bucket_name'])

        if 'prefix' in self.node_config:
            prefix = self.node_config['prefix']
        else:
            prefix = None

        blob_list = bucket.list_blobs(prefix=prefix)

        valid_blobs = [blob for blob in blob_list if self.node_config['blob_name'] in blob.name]

        if len(valid_blobs) == 0:
            raise Exception('{} not found in GCS bucket.'.format(self.node_config['blob_name']))

        return [blob.download_as_string() for blob in valid_blobs]

    def run(self, data_object):
        """Read serialized object(s) from GCS bucket which contain the blob_name

        Args:
            data_object: DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        logging.info('Reading {} from GCS'.format(self.node_config['blob_name']))

        if self.node_config['deserializer'] not in Deserializer.SUPPORTED_DESERIALIZERS.keys():
            logging.warning(f"{self.node_config['deserializer']} deserializer not supported.")
            logging.warning(f"The following deserializers are supported {Deserializer.SUPPORTED_DESERIALIZERS.keys()}.")
            raise Exception(f"Unsupported Deserializer: {self.node_config['deserializer']}")

        deserializer = Deserializer.SUPPORTED_DESERIALIZERS[self.node_config['deserializer']]

        objects = [deserializer.loads(obj) for obj in self.download_blobs_as_strings()]

        terminate = len(objects) == 0

        if len(objects) == 1:
            data_object.add(self, objects[0], key=GcsDeserializer.DATA_KEY)
        else:
            data_object.add(self, objects, key=GcsDeserializer.DATA_KEY)

        return data_object, terminate



import pytest
import os
import pandas as pd

from primrose.readers.postgres_reader import PostgresReader
from primrose.configuration.configuration import Configuration
from primrose.data_object import DataObject, DataObjectResponseType
from unittest.mock import patch

def test_necessary_config():
    assert len(PostgresReader.necessary_config({})) == 1


@pytest.mark.optional
@pytest.mark.postgres
def test_run(monkeypatch):
    config = {
        "implementation_config": {
            "reader_config": {
                "mynode": {
                    "class": "PostgresReader",
                    "query_json": [
                        {"query": "test/test_mysql.sql"}
                    ],
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    keys = ["POSTGRES_HOST",
            "POSTGRES_PORT",
            "POSTGRES_DB",
            "POSTGRES_USER",
            "POSTGRES_PASS"]
    for i, k in enumerate(keys):
        os.environ[k] = str(i)

    reader = PostgresReader(configuration, 'mynode')

    with patch('psycopg2.connect') as mock_connect:

        def fake_df(query, con):
            return pd.DataFrame({'Name':['Tom', 'nick', 'krish', 'jack'], 'Age':[20, 21, 19, 18]})
        monkeypatch.setattr(pd,'read_sql',fake_df)

        data_object, terminate = reader.run(data_object)

        assert not terminate

        dd = data_object.get('mynode', rtype=DataObjectResponseType.KEY_VALUE.value)
        assert 'query_0' in dd
        df = dd['query_0']
        assert list(df.T.to_dict().values())[0] == {'Name': "Tom", "Age": 20}




import pytest
from primrose.pipelines.dataframe_joiner import DataFrameJoiner
from primrose.base.transformer_sequence import TransformerSequence
from primrose.configuration.configuration import Configuration
from primrose.readers.csv_reader import CsvReader
import pandas as pd
from primrose.data_object import DataObject
from primrose.data_object import DataObjectResponseType


def test_init_pipeline():
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mypipeline"]
                }
            },
            "pipeline_config": {
                "mypipeline":{
                    "class": "DataFrameJoiner",
                    "join_key": ["first"],
                    "start_table": "myreader"
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    pipeline = DataFrameJoiner(configuration, "mypipeline")
    ts = pipeline.init_pipeline()
    assert isinstance(ts, TransformerSequence)

def test_transform():
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader_left": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mypipeline"]
                },
                "myreader_right": {
                    "class": "CsvReader",
                    "filename": "test/merge_right3.csv",
                    "destinations": ["mypipeline"]
                }
            },
            "pipeline_config": {
                "mypipeline":{
                    "class": "DataFrameJoiner",
                    "join_key": ["first"],
                    "start_table": "myreader_left",
                    "is_training": True
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    left_df = pd.read_csv("test/minimal.csv")
    reader_left = CsvReader(configuration, "myreader_left")
    data_object.add(reader_left, left_df)

    right_df = pd.read_csv("test/merge_right3.csv")
    reader_right = CsvReader(configuration, "myreader_right")
    data_object.add(reader_right, right_df)

    pipeline = DataFrameJoiner(configuration, "mypipeline")

    data_object, terminate = pipeline.run(data_object)

    assert not terminate

    joined_data = data_object.get("mypipeline", pop_data=True, rtype=DataObjectResponseType.VALUE.value)
    assert joined_data.shape[0] == 2

    assert list(joined_data.T.to_dict().values())[0] == {'first': 'joe', 'last': 'doe', 'age': 47}
    assert list(joined_data.T.to_dict().values())[1] == {'first': 'mary', 'last': 'poppins', 'age': 42}


def test_transform2():
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader_left": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mypipeline"]
                },
                "myreader_right": {
                    "class": "CsvReader",
                    "filename": "test/merge_right3.csv",
                    "destinations": ["mypipeline"]
                }
            },
            "pipeline_config": {
                "mypipeline":{
                    "class": "DataFrameJoiner",
                    "join_key": ["first"],
                    "start_table": "JUNK",
                    "is_training": True
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    left_df = pd.read_csv("test/minimal.csv")
    reader_left = CsvReader(configuration, "myreader_left")

    right_df = pd.read_csv("test/merge_right3.csv")
    reader_right = CsvReader(configuration, "myreader_right")

    #note: am deliberately swapping order to right is first
    data_object.add(reader_right, right_df)
    data_object.add(reader_left, left_df)

    pipeline = DataFrameJoiner(configuration, "mypipeline")

    with pytest.raises(Exception) as e:
        pipeline.run(data_object)
    assert "Could not find start_table in upstream keys: JUNK" in str(e)



from primrose.base.node import AbstractNode

class TestExtNode(AbstractNode):
    
    def necessary_config(self, node_config):
        return ([])

    def run(self, data_object):
        terminate = False
        return data_object, terminate



"""Module with AbstractNode implementation, Example for primrose user modification

Author(s):
    Mike Skarlinski (michael.skarlinski@ww.com)

"""
from primrose.base.reader import AbstractReader


class AwesomeReader(AbstractReader):
    """(EXAMPLE READER) Read input directly from the configuration file"""

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the AwesomeReader object

        Put your required configuration keys here

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            filename: name of the file

        Returns:
            set of necessary keys for the AwesomeReader object

        """
        return set(['data_to_read'])

    def run(self, data_object):
        """Read data from node_config

        Returns:
            data_object (DataObject): DataObject instance
            terminate (bool): should we terminate the DAG? true or false

        """

        # Example showing data being created directly from the config file
        # any valid python can be used here for reading data into the data_object
        # examples of other readers can be found in primrose/readers/*
        # -------------------------------
        data = self.node_config['data_to_read']
        print('Reading in data!: {}'.format(data))
        # -------------------------------
        # add data into data object for use downstream
        data_object.add(self, data)
        terminate = data is None
        return data_object, terminate



'''A class that creates a directed acyclic graph (DAG) and perhaps a number of checks,
such as detecting cycles, orphans, and unrecognized nodes

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

'''
import logging
import networkx as nx
from primrose.configuration.util import OperationType, ConfigurationError
import matplotlib.pyplot as plt
from networkx.algorithms.shortest_paths.generic import has_path
from primrose.node_factory import NodeFactory
from primrose.base.conditional_path_node import AbstractConditionalPath


class ConfigurationDag():

    def __init__(self, config):
        """instantiate ComfigurationDAG instance

        Args:
            config (Configuration): Configuration instance

        """
        self.config = config

    @staticmethod
    def check_node_exists(node_names, key):
        """check that some specified destination is node on graph

        Args:
            nodes_names (list): list of node names
            key (str): name of node to check

        Raises:
            ConfigurationError

        """
        if not key in node_names:
            raise ConfigurationError("Destination %s does not exist" % key)

    @staticmethod
    def add_edge(G, G2, node_names, key, destination):
        """add an edge to the DAG

        Args:
            G (networkx bidirectional graph): bidirectional graph instance
            G2 (networkx directional graph): bidrectional graph instance
            key (str): starting node name
            destination (str): destination node name

        Returns:
            nothing. Side effect is to add the edge

        """
        ConfigurationDag.check_node_exists(node_names, destination)
        logging.debug("Adding edge %s -> %s" % (key, destination))
        G.add_edge(key, destination)
        G2.add_edge(key, destination)

    def starting_nodes(self):
        """Where does the DAG start? Compute list of starting (level 0) nodes

        Returns:
            list of node name

        """
        # Level 0 nodes in a directed graph will have 1 or more out_edges but no in_edges
        nodes_with_outs = set(e[0] for e in self.G2.out_edges())
        nodes_with_ins = set(e[1] for e in self.G2.in_edges())
        return nodes_with_outs - nodes_with_ins

    def nodes_of_type(self, operation_type):
        """get set of nodes of a given operation type (OperationType.reader, OperationType.writer etc)

        Args:
            operation_type (OperationType): OperationType

        Returns:
            set of keys, if any, of the given operation type

        """
        return set([k for k in self.node_map.keys() if self.node_map[k] == operation_type.value])

    def upstream_nodes_of_type(self, target_node_name, operation_type):
        """get set of nodes of a given operation type (OperationType.reader, OperationType.writer etc)
            upstream of some given target node

        Args:
            operation_type (OperationType): OperationType

        Returns:
            set of keys, if any, of the given operation type

        """
        assert target_node_name in self.node_map
        nodes = self.nodes_of_type(operation_type)
        nodes = [node for node in nodes if has_path(self.G2, node, target_node_name)]
        return set(nodes)

    def descendents(self, source):
        """Get the list of descendents from source, i.e. subgraph below source

        Args:
            source (str): name of source

        Returns:
            list of descendents of source node

        """
        assert source in self.node_map
        return nx.descendants(self.G2, source)

    def paths(self, source, target):
        """return the paths, if any, from a given source node to a given target node

        Args:
            source (str): name of node which is starting point of path
            target (str): name of node which is end point of path

        Returns:
            list of list of nodes (in order) forming the paths, or None if no path

        """
        assert source in self.node_map
        assert target in self.node_map
        if has_path(self.G2, source, target):
            return nx.all_simple_paths(self.G2, source=source, target=target)
        return None

    def check_for_cycles(self):
        """check for cycles

        Raise:
            ConfigurationError if cycles found

        """
        cycles = None
        try:
            cycles = nx.find_cycle(self.G2)
            logging.info("cycles:" + str(cycles))
        except nx.exception.NetworkXNoCycle as e:
            # it throws an error if there are no cycles, the opposite of what we want
            if str(e) != "No cycle found.":
                raise e # pragma: no cover
            else:
                logging.info("OK: no cycles found")
        #to get here, we must have cycles!
        if cycles:
            raise ConfigurationError("Cycle(s) found: %s" % str(cycles))

    def check_connected_components(self):
        """now we can count the number of connected components. >1 is problem

        Raises:
            ConfigurationError if multiuple connected components

        """
        connected_components = nx.connected_components(self.G)
        n = sum([1 for c in connected_components])
        if n > 1:
            raise ConfigurationError("Found multiple connected components: %s" % str(list( nx.connected_components(self.G) )))
        else:
            logging.info("OK: 1 connected component")

    def upstream_keys(self, instance_name):
        """get list of keys (names of nodes in the DAG) that feed into instance_name node

        Args:
            instance_name (str): name of instance

        Returns:
            list of nodes

        """
        if not self.G2.has_node(instance_name):
            raise ConfigurationError("Node not found in the DAG: %s" % instance_name)
        return list(self.G2.predecessors(instance_name))

    def upstream_typed_keys(self, instance_name):
        """get dictionary of the upstream keys with Operation types as values

        Args:
            instance_name (str): name of instance

        Returns:
            dictionary of {name : node type}

        """
        keys = self.upstream_keys(instance_name)
        d = {}
        for k in keys:
            d[k] = self.node_map[k]
        return d

    def create_dag(self):
        """Create the DAG

        Returns:
            nothing. Side effect is to set up graphs and node map

        """
        logging.info("Checking configuration DAG")
        G = nx.Graph()
        G2 = nx.DiGraph()
        node_names = set()
        cleanup_nodes = set()

        some_postprocess_node = None

        # key to section type
        node_map = {}

        self.conditional_nodes = set()

        #add the nodes to the graph:
        for section_key in self.config.keys():

            for key in self.config[section_key].keys():
                logging.debug("Adding node '%s'" % key)
                G.add_node(key)
                G2.add_node(key)
                node_names.add(key)

                node_map[key] = section_key

                # root out conditional nodes...
                node_config = self.config[section_key][key]
                node_class = node_config['class']
                class_obj = NodeFactory().name_dict[node_class]
                if issubclass(class_obj, AbstractConditionalPath):
                    self.conditional_nodes.add(key)

                # cleanup section can be disconnected from rest of graph so let's keep track of these nodes
                if section_key == OperationType.cleanup.value:
                    cleanup_nodes.add(key)

                # hack: we are going to add an edge from a postprcocess step (any one) to cleanup nodes so they
                # are not a separate connected component
                if section_key == OperationType.postprocess.value and some_postprocess_node is None:
                    some_postprocess_node = key

        # add the edges
        for section_key in self.config.keys():

            for key in self.config[section_key].keys():
                d = self.config[section_key][key]

                if 'destinations' in d:
                    for destination in d['destinations']:

                        if not isinstance(destination, str):
                            raise ConfigurationError("Unrecognized destination type: %s" % destination)

                        if destination in node_map:
                            ConfigurationDag.add_edge(G,G2,node_names,key,destination)
                        else:
                            raise ConfigurationError("Did not find %s destination in %s.%s" % (destination, section_key, key))

        logging.info("OK: good referential integrity")

        self.G = G
        self.G2 = G2
        self.node_map = node_map

    def check_dag(self):
        """check that it is a DAG

        Note:
            check for cycles
            check only 1 connected component, no orphans
            check that all edges point to known nodes

        Raises:
            Excetions if cycles found or multiple connected components

        """
        self.create_dag()

        self.check_connected_components()

        self.check_for_cycles()

    def plot_dag(self, filename, traverser, node_size=500, label_font_size=12, text_angle=0, image_width=16, image_height=12):
        """plot the DAG to image file

        Args:
            filename (str): path to write image to
            title (str): title to add to chart
            node_size (int): node size
            label_font_size (int): font size
            text_angle (int): angle to rotate. This is angle in degrees counter clockwise from east
            image_width (int): width of image in inches
            image_height (int): heightof image in inches

        Returns:
            nothing. Saves image to file

        """
        # map nodes to a color for their operation type
        # https://stackoverflow.com/questions/27030473/how-to-set-colors-for-nodes-in-networkx-python
        color_map = []
        colors = ['#fbb4ae','#b3cde3','#ccebc5','#decbe4','#fed9a6']
        for node in self.G2:
            if self.node_map[node] == OperationType.reader.value:
                color_map.append(colors[0])
            elif self.node_map[node] == OperationType.pipeline.value:
                color_map.append(colors[1])
            elif self.node_map[node] == OperationType.model.value:
                color_map.append(colors[2])
            elif self.node_map[node] == OperationType.writer.value:
                color_map.append(colors[3])
            else:
                color_map.append(colors[4])

        fig = plt.figure(figsize=(image_width,image_height))
        ax = plt.subplot(111)
        ax.set_title(filename, fontsize=10)

        try:
            import pydot
            from networkx.drawing.nx_pydot import graphviz_layout
        except ImportError: # pragma: no cover
            raise ImportError(
                "This example needs Graphviz and pydot."
                "Please refer to the Plotting requirements in the README"
            )

        # pos = nx.spring_layout(G)
        # pos = nx.circular_layout(G)
        # pos = nx.kamada_kawai_layout(G)
        # pos = nx.shell_layout(G)
        # pos = nx.spectral_layout(G)
        pos = graphviz_layout(self.G2, prog='dot') #, prog='twopi', args='')

        nx.draw(self.G2, pos, node_size=node_size,node_color = color_map, edge_color='#939393', font_size=8, font_weight='bold')
        # nx.draw_networkx_nodes(G, pos, node_color='b', node_size=500, alpha=0.8)

        if len(self.conditional_nodes) > 0:
            cnodes = nx.draw_networkx_nodes(self.G2, pos, node_color='#e6b655', node_size=1.5*node_size, alpha=0.8, node_shape="D", nodelist=list(self.conditional_nodes))
            cnodes.set_edgecolor('red')

#        nx.draw_networkx_labels(self.G2,pos, font_size=9)

        text = nx.draw_networkx_labels(self.G2,pos,with_labels=False, font_size=label_font_size) #, bbox=Bbox.from_bounds(x0, y0, 20,30)) #, wrap=True)

        if traverser:
            #map node name to sequence number
            sequence = traverser.traversal_list()
            idx = list(range(1,len(sequence) + 1))
            d = dict(zip(sequence,idx))

            # let's plot the sequence numner above the node. How far above it?
            ys = [t._y for _,t in text.items()]
            ysrange = max(ys) - min(ys)
            offset = 0.02 * abs(ysrange)

        for _,t in text.items():
            t.set_rotation(text_angle)

            if traverser:
                plt.text(t._x, t._y + offset, d[t._text], fontsize=24, color='red')

        plt.axis('off')
        plt.tight_layout()
        plt.savefig(filename, format="PNG")
        logging.info("Graph written to %s" % filename)



"""Module to traverse DAG by running all readers, then all pipelines, then all models, 
then all postprocess, then all cleanup

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.dag.dag_traverser import DagTraverser
from primrose.dag.depth_first_traverser import DepthFirstTraverser

class ConfigLayerTraverser(DagTraverser):
    """Module to traverse DAG by running all readers, then all pipelines,
    then all models, then all postprocess, then all cleanup

    """

    def _sort_by_depth_first(self, candidates):
        depth_first_order = DepthFirstTraverser(self.configuration).traversal_list()
        return [n for n in depth_first_order if n in candidates]

    def run_section_by_section(self):
        """Do we want go through section by section?

        Note:
            suppose we had 5 sections and we asked to run just 2: section1 and section2.
            Is it important that we run section1 and then section 2 (if so, return True)
            or just that we run all the nodes from section1 and section2 together in some
            desired order (return False)

        Returns:
            result (Boolean)

        """
        return True

    def traversal_list(self):
        """get list of nodes to traverse doing all readers, then all pipelines, then all models etc,
        or from user-defined list from metadata.section_registry

        Returns:
            sequence (list): list of node names in the order that they will be run

        """
        sequence = []
        sections, source = self.configuration.sections_in_order()
        for section_name in sections:

            section_keys = self.configuration.config[section_name].keys()

            sorted_keys = self._sort_by_depth_first(section_keys)

            sequence.extend(sorted_keys)

        return sequence



"""Module to run a basic decision tree model

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""
import pandas as pd
import numpy as np
import logging
from sklearn import preprocessing

from primrose.base.transformer import AbstractTransformer


class ExplicitCategoricalTransform(AbstractTransformer):

    DEFAULT_NUMERIC = -9999

    def __init__(self, categoricals):
        """initialize the ExplicitCategoricalTransform
        Args:
            categoricals: dictionary containing for each column to be transformed:
            - transformations: list of strings to be executed on the data ('x' represents the current categorical variable)
            - rename: if present, rename the current categorical variable to that name
            - to_numeric: if true, attempt to apply to_numeric after previous transformations

        """
        self.categoricals = categoricals

    def fit(self, data):
        pass

    @staticmethod
    def _process_transformations(data, input_data, categorical, x):
        """transform a column

        Args:
            data (dataframe): dataframe

            input configuration (JSON): JSON categorical config for this variable

            categorical (str): varible name

            x (str): transformation string

        Returns:
            data (dataframe)

        """
        if 'transformations' in input_data.keys():
            logging.info("Applying key {} to variable {}".format('transformations', categorical))
            for transformation in input_data['transformations']:
                exec(transformation.format(x=x))

    @staticmethod
    def _process_rename(data, input_data, categorical):
        """rename a field

        Args:
            data (dataframe): dataframe

            input configuration (JSON): JSON categorical config for this variable

            categorical (str): varible name

        Returns:
            (tuple): tuple containing:

                data (dataframe): dataframe

                name (str): original name (if not "to_numeric": True), new_name otherwise
        """
        if 'rename' in input_data.keys():
            logging.info("Applying key {} to variable {}".format('rename', categorical))
            data = data.rename({categorical: input_data['rename']}, axis='columns')
            return data, input_data['rename']
        return data, categorical

    @staticmethod
    def _process_numeric(data, input_data, name):
        """convert column to numeric

        Args:
            data (dataframe): dataframe

            input configuration (JSON): JSON categorical config for this variable

            name (str): field name

        Returns:
            data with the colun converted to numeric

        """
        if input_data.get('to_numeric', False):

            logging.info("Applying key {} to variable {}".format('to_numeric', name))

            # if there are errors converting to numerical values, we need to sub in a reasonable value
            if sum(pd.to_numeric(data[name], errors='coerce').isnull()) > 0:
                logging.info("Can't convert these entries in {}. Replacing with {}: {}".format(
                    name, ExplicitCategoricalTransform.DEFAULT_NUMERIC,
                    np.unique(data[name][pd.to_numeric(data[name], errors='coerce').isnull()].astype(str))))

                data[name][pd.to_numeric(data[name], errors='coerce').isnull()] = ExplicitCategoricalTransform.DEFAULT_NUMERIC
            try:
                data[name] = pd.to_numeric(data[name])
                return data

            except:
                raise TypeError('Failed to convert feature {} to numeric'.format(name))

        else:
            return data

    def transform(self, data):
        """Transform categorical variables into one or more numeric ones, no need to separate testing & training data

        Args:
            data: dictionary containing dataframe with all categorical columns present

        Returns:
            data with all categorical columns recoded and/or deleted

        """
        for categorical in self.categoricals.keys():

            x = "data['{}']".format(categorical)

            input_data = self.categoricals[categorical]

            ExplicitCategoricalTransform._process_transformations(data, input_data, categorical, x)

            data, new_name = ExplicitCategoricalTransform._process_rename(data, input_data, categorical)

            data = ExplicitCategoricalTransform._process_numeric(data, input_data, new_name)

        return data


class ImplicitCategoricalTransform(AbstractTransformer):
    """Class which implicitly transforms all string columns of a dataframe with sklearn LabelEncoder"""

    def __init__(self, target_variable):
        """initialize this ImplicitCategoricalTransform

        Args:
            target_variable (str): target variable name

        """
        self.target_variable = target_variable
        self._encoder = {}
        self.target_encoder = None

    def fit(self, data):
        """encode the data as categorical labels

        Args:
            data (dataframe)

        Returns:
            dataframe (dataframe)

        """

        logging.info('Fitting LabelEncoders on all string-based dataframe columns...')

        data.is_copy = False

        for column_name in data.columns:

            if data[column_name].dtype == object:

                logging.info('Fitting LabelEncoder for column {}'.format(column_name))

                self._encoder[column_name] = preprocessing.LabelEncoder()
                self._encoder[column_name].fit(data[column_name])

                if column_name == self.target_variable:
                    self.target_encoder = self._encoder[column_name]
            else:
                pass

        return data

    def transform(self, data):
        """Transform data into categorical variables using pre-trained label encoder

        Args:
            data (dataframe)

        Returns:
            dataframe (dataframe)

        """

        data.is_copy = False

        for column_name in data.columns:

            if column_name in self._encoder:

                logging.info('LabelEncoding column {}'.format(column_name))

                data[column_name] = self._encoder[column_name].transform(data[column_name])

        return data



"""Transformer that wraps around pandas.Series.str methods

Author(s):
    Brian Graham (brian.graham@weightwatchers.com)

"""

import pandas as pd
import logging
from primrose.base.transformer import AbstractTransformer

class StringTransformer(AbstractTransformer):
    """Transforms Series of strings in a Series or DataFrame."""
    def __init__(self,method,columns,*args,**kwargs):
        """

        Args:
            method (str): pandas.Series.str method
            columns (str or list): single column name str or list of columns to operate on
            *args: args for given string method
            **kwargs: kwargs for given string method

        """
        self.method = method
        self.columns = columns
        self.args = args
        self.kwargs = kwargs

    def fit(self, df):
        """fit data, here just passing.

        Args:
            data (object): some data

        """
        pass

    def transform(self, df):
        """Applies string operation on dataframe columns.

        Args:
            df (pd.DataFrame): pandas dataframe

        Returns:
            df (pd.DataFrame): pandas dataframe

        """
        if isinstance(self.columns,str):
            df[self.columns] = self._execute_str_method(df[self.columns])
        elif isinstance(self.columns,list):
            for col in self.columns:
                df[col] = self._execute_str_method(df[col])
        else:
            logging.info('Column not passed as string or list of columns, returning original dataframe.')
        return df

    def _execute_str_method(self,series):
        """Executes string method on pandas series.

        Args:
            series (pd.Series): pandas series to transform

        Returns:
            series (pd.Series): transformed pandas series

        """
        method = getattr(series.str, self.method)
        return method(*self.args, **self.kwargs)




import pytest
import sys
import os
from primrose.data_object import DataObject
from primrose.data_object import DataObjectResponseType
from primrose.configuration.configuration import Configuration
from primrose.configuration.util import OperationType
from primrose.readers.csv_reader import CsvReader

@pytest.fixture()
def setup_vars():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_s3_writer']
                }
            },
            "writer_config": {
                "recipe_s3_writer": {
                    "class": "S3Writer",
                    "dir": "cache",
                    "key": "data",
                    "bucket_name": "does_not_exist_bucket_name",
                    "bucket_filename": "does_not_exist.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)
    return configuration, data_object

def test_repr(setup_vars):
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    assert str(data_object) == "DataObject:defaultdict(<class 'dict'>, {'csv_reader': {'data': 'TESTING'}})"

def test_add(setup_vars):
    '''test value'''
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    response = data_object.get('csv_reader',rtype=DataObjectResponseType.VALUE.value)
    assert response == data_to_save

def test_add2(setup_vars):
    '''test key value'''
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    response = data_object.get('csv_reader',rtype=DataObjectResponseType.KEY_VALUE.value)
    assert isinstance(response, dict)
    assert DataObject.DATA_KEY in response
    assert response[DataObject.DATA_KEY] == data_to_save

def test_add3(setup_vars):
    '''test instance key value'''
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    response = data_object.get('csv_reader',rtype=DataObjectResponseType.INSTANCE_KEY_VALUE.value)
    assert isinstance(response, dict)
    assert 'csv_reader' in response
    assert 1 == len(list(response.keys()))
    assert response['csv_reader'][DataObject.DATA_KEY] == data_to_save

def test_add4(setup_vars):
    '''test add 2 items'''
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    data_object.add(reader, data_to_save, overwrite=True)

    with pytest.raises(Exception) as e:
        data_object.add(reader, data_to_save, overwrite=False)
    assert "Key already exists for csv_reader" in str(e)

def test_get(setup_vars):
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    with pytest.raises(Exception) as e:
        data_object.get('reader', rtype='junk')
    assert "Unrecognized rtype: junk" in str(e)

def test_get2(setup_vars):
    configuration, data_object = setup_vars
    with pytest.raises(Exception) as e:
        data_object.get('junk')
    assert "Key not found: junk" in str(e)

def test_get3(setup_vars):
    configuration, data_object = setup_vars
    assert len(data_object.data_dict.keys()) == 0

    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)
    assert len(data_object.data_dict.keys()) == 1

    data_object.get('csv_reader',pop_data=False)
    assert len(data_object.data_dict.keys()) == 1

    data_object.get('csv_reader',pop_data=True)
    assert len(data_object.data_dict.keys()) == 0

def test_get4(setup_vars):
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_object.add(reader, key='k1', data='v1')
    data_object.add(reader, key='k2', data='v2')
    response = data_object.get(reader.instance_name, rtype=DataObjectResponseType.VALUE.value)
    assert isinstance(response, dict)
    assert len(response.keys()) == 2

def test_upstream_keys(setup_vars):
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)

    keys = data_object.upstream_keys('recipe_s3_writer')
    assert keys == ['csv_reader']

    keys = data_object.upstream_keys('recipe_s3_writer', operation_type_filter=OperationType.reader)
    assert keys == ['csv_reader']

    keys = data_object.upstream_keys('recipe_s3_writer', operation_type_filter=OperationType.writer)
    assert keys == []

def test_get_upstream_data(setup_vars):
    configuration, data_object = setup_vars
    with pytest.raises(Exception) as e:
        data_object.get_upstream_data('recipe_s3_writer')
    assert "No upstream keys with data found for recipe_s3_writer" in str(e)

def test_get_upstream_data2(setup_vars):
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)
    response = data_object.get_upstream_data('recipe_s3_writer')
    assert isinstance(response, dict)
    assert DataObject.DATA_KEY in response
    assert response[DataObject.DATA_KEY] == data_to_save

def test_get_upstream_data3(setup_vars):
    configuration, data_object = setup_vars
    reader = CsvReader(configuration, 'csv_reader')
    data_to_save = "TESTING"
    data_object.add(reader, data_to_save)
    response = data_object.get_upstream_data('recipe_s3_writer')
    assert isinstance(response, dict)
    assert DataObject.DATA_KEY in response
    assert response[DataObject.DATA_KEY] == data_to_save

def test_get_upstream_data4():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader1": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_s3_writer']
                },

                "csv_reader2": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_s3_writer']
                }
            },
            "writer_config": {
                "recipe_s3_writer": {
                    "class": "S3Writer",
                    "dir": "cache",
                    "key": "data",
                    "bucket_name": "does_not_exist_bucket_name",
                    "bucket_filename": "does_not_exist.csv"
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    reader1 = CsvReader(configuration, 'csv_reader1')
    reader2 = CsvReader(configuration, 'csv_reader2')

    data_object.add(reader1, "data1")
    data_object.add(reader2, "data2")

    response = data_object.get_upstream_data('recipe_s3_writer')
    assert isinstance(response, dict)
    assert 'csv_reader1' in response
    assert 'csv_reader2' in response
    assert response['csv_reader1'][DataObject.DATA_KEY] == "data1"
    assert response['csv_reader2'][DataObject.DATA_KEY] == "data2"

    response = data_object.get_upstream_data('recipe_s3_writer')
    assert isinstance(response, dict)
    assert response['csv_reader1'][DataObject.DATA_KEY] == "data1"
    assert response['csv_reader2'][DataObject.DATA_KEY] == "data2"

def test_caching():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    writer = CsvReader(configuration, "csv_reader")

    data_object.add(writer, "some_data")

    filename = "test_data_object_cache.pkl"
    if os.path.exists(filename):
        os.remove(filename)

    data_object.write_to_cache(filename)

    assert os.path.exists(filename)

    restored_data_object = DataObject.read_from_cache(filename)

    assert isinstance(restored_data_object, DataObject)

    assert restored_data_object.get("csv_reader", rtype=DataObjectResponseType.VALUE.value) == "some_data"

    if os.path.exists(filename):
        os.remove(filename)

def test_get_filtered_upstream_data():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_s3_writer']
                }
            },
            "writer_config": {
                "recipe_s3_writer": {
                    "class": "S3Writer",
                    "dir": "cache",
                    "key": "data",
                    "bucket_name": "does_not_exist_bucket_name",
                    "bucket_filename": "does_not_exist.csv"
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    reader = CsvReader(configuration, 'csv_reader')

    data_object.add(reader, "some_data_to_save")

    data = data_object.get_filtered_upstream_data("recipe_s3_writer", "data")
    assert data == {'data': 'some_data_to_save'}
    assert not isinstance(data, list)

    data = data_object.get_filtered_upstream_data("recipe_s3_writer", "JUNK")
    assert not data


def test_get_filtered_multiple_upstream_data():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader1": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_s3_writer']
                },
                "csv_reader2": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['recipe_s3_writer']
                }
            },
            "writer_config": {
                "recipe_s3_writer": {
                    "class": "S3Writer",
                    "dir": "cache",
                    "key": "data",
                    "bucket_name": "does_not_exist_bucket_name",
                    "bucket_filename": "does_not_exist.csv"
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    reader1 = CsvReader(configuration, 'csv_reader1')
    reader2 = CsvReader(configuration, 'csv_reader2')

    data_object.add(reader1, "some_data_to_save")
    data_object.add(reader2, "some_data_to_save")

    data = data_object.get_filtered_upstream_data("recipe_s3_writer", "data")
    assert data == [{'data': 'some_data_to_save'}, {'data': 'some_data_to_save'}]
    assert isinstance(data, list)

    data = data_object.get_filtered_upstream_data("recipe_s3_writer", "JUNK")
    assert not data


"""Module with AbstractNode implementation, able to read from CSV

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""

import logging
import pandas as pd
from primrose.base.reader import AbstractReader

class CsvReader(AbstractReader):
    """Reads CSV file into a pandas dataframe"""

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the CsvReader object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            filename: name of the file

        Returns: 
            set of necessary keys for the CsvReader object

        """
        return set(['filename'])

    def get_optional_config(self):
        """Optionally get kwargs to pass to pandas csv reader.

        Notes:
            kwargs (dict): dictionary of kwargs key-value pairs

        Example:
            "csv_reader": { \
                "class": "CsvReader", \
                "filename": "data/mydata.csv", \
                "kwargs": { \
                "header": None, \
                "sep": ":" \
                } \
            } \

        """
        if 'kwargs' in self.node_config:
            return self.node_config['kwargs']
        else:
            return {}

    def run(self, data_object):
        """Read CSV to a pandas dataframe

        Returns:
            data_object (DataObject): DataObject instance
            terminate (bool): should we terminate the DAG? true or false

        """
        filename = self.node_config['filename']
        logging.info('Reading {} from CSV'.format(filename))
        kwargs = self.get_optional_config()
        df = pd.read_csv(filename, **kwargs)
        data_object.add(self, df)
        terminate = df.empty
        return data_object, terminate



"""Module with AbstractNode implementation, able to read from GCS

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""

import logging
import dill
from google.cloud import storage
import warnings

from primrose.base.reader import AbstractReader


class GcsDillReader(AbstractReader):
    """Read a file from Gcs and un-dills it into memory"""

    DATA_KEY = 'reader_data'
    warnings.warn('Use Deserializer instead. GcsDillReader will be deprecated in a future release.', DeprecationWarning)

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the GcsDillReader object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            bucket_name: name of the GCS bucket
            blob_name: name of the blob

        Returns:
            set of necessary keys for the CsvReader object

        """
        return set(['bucket_name', 'blob_name'])

    def download_blobs_as_strings(self):
        """Downloads a blob from the bucket contining the user specified blob_name

        Returns:
            list of strings

        """

        if 'gcs_project' in self.node_config:
            storage_client = storage.Client(project=self.node_config['gcs_project'])
        else:
            storage_client = storage.Client()

        bucket = storage_client.get_bucket(self.node_config['bucket_name'])

        if 'prefix' in self.node_config:
            prefix = self.node_config['prefix']
        else:
            prefix = None

        blob_list = bucket.list_blobs(prefix=prefix)

        valid_blobs = [blob for blob in blob_list if self.node_config['blob_name'] in blob.name]

        if len(valid_blobs) == 0:
            raise Exception('{} not found in GCS bucket.'.format(self.node_config['blob_name']))

        return [blob.download_as_string() for blob in valid_blobs]

    def run(self, data_object):
        """Read dill object(s) from GCS bucket which contain the blob_name

        Args:
            data_object: DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        logging.info('Reading {} from GCS'.format(self.node_config['blob_name']))
        objects = [dill.loads(obj) for obj in self.download_blobs_as_strings()]

        terminate = len(objects) == 0

        if len(objects) == 1:
            data_object.add(self, objects[0], key=GcsDillReader.DATA_KEY)
        else:
            data_object.add(self, objects, key=GcsDillReader.DATA_KEY)

        return data_object, terminate



"""Module to run train test splits

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import logging
import pandas as pd
from sklearn.model_selection import train_test_split

from primrose.base.pipeline import AbstractPipeline, PipelineModeType
from primrose.data_object import DataObject, DataObjectResponseType


class TrainTestSplit(AbstractPipeline):
    """Parent pipeline to split into training / testing data, and run a transform

    Note:
        This class will split data into testing and training portions, then write the objects to the data_object
        if this is the only the pipeline operation needed, then you can use this class directly as a pipeline.
        Otherwise, make a child class and write an init_pipeline method to perform operations on your data.

        This class also handles writing your transformer_sequence into the data_object, so there's no need
        to write in child classes.

    """

    @staticmethod
    def necessary_config(node_config):
        """Return the necessary configuration keys for the DataFrameJoiner object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            target_variable (string): column name which holds the target variable \
            training_fraction (float): 0->1 float for the fraction of data rows to be used for training \
            seed (int): random number to control the stochastic row selection \

        Returns:
            set of keys

        """
        return set(['training_fraction', 'seed'])

    def features(self, data):
        """Use user-specified features if available, otherwise use all non-target columns
        
        Args:
            data (DataFrame): pandas dataframe

        Returns:
            lsit of feature names
        
        """
        if 'features' in self.node_config:
            return self.node_config['features']

        else:
            if 'target_variable' in self.node_config:
                return [f for f in data.columns if f != self.node_config['target_variable']]
            else:
                return data.columns

    def _train_test_split(self, data):
        """Split data into test/train sets
        
        Returns:
            train_data_to_transform (DataFrame)
            
            test_data_to_transform (DataFrame)
        
        """
        logging.info("Splitting data into testing and training sets.")

        test_size = (1.0 - float(self.node_config['training_fraction']))

        if test_size == 0:
            train_data_to_transform = data
            test_data_to_transform = pd.DataFrame()
        
        else:
            if 'target_variable' in self.node_config:
                data_train, data_test, target_train, target_test = train_test_split(
                    data[sorted(list(set(data.columns) - set([self.node_config.get('target_variable')])))],
                    data[self.node_config['target_variable']],
                    test_size=(1.0 - float(self.node_config['training_fraction'])),
                    random_state=self.node_config['seed'])

                # re-merge training and target data into a single dataframe for transforming
                train_data_to_transform = pd.concat([data_train, target_train], axis=1)
                test_data_to_transform = pd.concat([data_test, target_test], axis=1)

            else:
                data_train, data_test = train_test_split(
                    data[sorted(list(set(data.columns) - set([self.node_config.get('target_variable')])))],
                    test_size=(1.0 - float(self.node_config['training_fraction'])),
                    random_state=self.node_config['seed'])

                train_data_to_transform = data_train
                test_data_to_transform = data_test

        logging.info('Training data rows: {}, Testing data rows: {}'.format(len(train_data_to_transform),
                                                                            len(test_data_to_transform)))

        return train_data_to_transform, test_data_to_transform

    def final_data_object_additions(self, data_object):
        """DataObject: Template method to be overloaded in child classes

        Note:
            Allows for child class TransformerSequence objects to be added into the data_object

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """
        return data_object

    @staticmethod
    def _concatenate_upstream_dataframes(data):
        """Concatenate multiple upstream data sources into a single dataframe

        Note:

            Upstream objects must add dataframes with the same schema, otherwise mismatched columns will have
            all NULL values in the mutual exclusive indexes

        Args:
            data (list): list of dicts keyed to instances

        Returns:
            dataframe (DataFrame): concatenated dataframes from the data list

        """

        dataframes_to_join = []

        for source in data:

            for key in data[source]:

                if isinstance(data[source][key], pd.DataFrame):

                    if dataframes_to_join:
                        if set(data[source][key].columns) != set(dataframes_to_join[0].columns):
                            logging.warning('Concatenated dataframe schemas do not match, subbed with NULL values.')

                    dataframes_to_join.append(data[source][key])

        return pd.concat(dataframes_to_join).reset_index(drop=True)

    def fit_transform(self, data_object):
        """Split data into testing and training sets, then applies the categorical transform to each

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """

        # we're can expect multiple objects from a reader, so we need to concatenate
        data_list = data_object.get_upstream_data(self.instance_name,
                                            pop_data=False,
                                            rtype=DataObjectResponseType.INSTANCE_KEY_VALUE.value)

        data = self._concatenate_upstream_dataframes(data_list)

        train_data, test_data = self._train_test_split(data)

        if not train_data.empty:
            train_data = self.execute_pipeline(train_data, PipelineModeType.FIT_TRANSFORM)
            data_object.add(self, train_data[self.features(train_data)], key='data_train', overwrite=False)

            if 'target_variable' in self.node_config:
                data_object.add(self, train_data[self.node_config['target_variable']], key='target_train',
                            overwrite=False)

        # run the pre-trained pipeline on the testing data
        if not test_data.empty:
            # run the pipeline in Transform mode since we've already fit the pipeline with training data
            test_data = self.execute_pipeline(test_data, PipelineModeType.TRANSFORM)
            self.data = test_data  # assign the data to the testing data if available
            data_object.add(self, test_data[self.features(train_data)], key='data_test', overwrite=False)

            if 'target_variable' in self.node_config:
                data_object.add(self, test_data[self.node_config['target_variable']], key='target_test',
                            overwrite=False)

        # save pipeline for writing later
        data_object.add(self, self.transformer_sequence, key='transformer_sequence', overwrite=False)
        data_object = self.final_data_object_additions(data_object)

        return data_object

    def transform(self, data_object):
        """Transform the data into label encoded data using the pre-trained transformer sequence

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """
        # we're can expect multiple objects from a reader, so we need to concatenate
        data_list = data_object.get_upstream_data(self.instance_name,
                                                pop_data=False,
                                                rtype=DataObjectResponseType.INSTANCE_KEY_VALUE.value)

        data = self._concatenate_upstream_dataframes(data_list)

        data = self.execute_pipeline(data, PipelineModeType.TRANSFORM)

        self.data = data  # keep the data for use in the final_data_object_additions method

        data_object.add(self, data[self.features(data)], key='data_test', overwrite=False)

        if 'target_variable' in self.node_config:
            data_object.add(self, data[self.node_config['target_variable']], key='target_test', overwrite=False)

        data_object = self.final_data_object_additions(data_object)

        return data_object




import pytest
from primrose.base.transformer import AbstractTransformer
import logging
from testfixtures import LogCapture

def test_fit_transform():
    class TestTransformer(AbstractTransformer):
        def fit(self, data):
            logging.info("Transfer FIT CALLED")

        def transform(self, data):
            logging.info("Transfer TRANSFORM CALLED")
            return data

    t = TestTransformer()

    with LogCapture() as l:
        t.fit_transform(None)
    l.check(
        ('root', 'INFO', 'Transfer FIT CALLED'),
        ('root', 'INFO', 'Transfer TRANSFORM CALLED')
    )




"""Module with AbstractReader implementation, able to read datasets from R

    This is a simple example to show that primrose can work with other languages.
    Note:
        It works with simple 2D R dataframes (e.g. iris dataset) but not higher tensors (e.g. iris3 dataset)
        It does not work with timeseries, such as `lh` dataset

        We perform an R as.data.frame conversion on all data to help stadardize the data structure.
        For instance, the R `euro` dataset is essentially a labelled vector but converts nicely to a dataframe.

Author(s):
    Carl Anderson (carl.anderson@ww.com)

"""
from primrose.base.reader import AbstractReader
import logging
import pandas as pd

class RReader(AbstractReader):
    """Reads in canned dataset from R"""

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            dataset (string): name of dataset in R

        Returns:
            set of keys necessary to run implementation

        """
        return set(['dataset'])

    def run(self, data_object):
        """Read canned dataset from R to a pandas dataframe

        Returns:
            data_object (DataObject): DataObject instance
            terminate (bool): should we terminate the DAG? true or false

        """
        dataset = self.node_config['dataset']
        logging.info('Reading {} from R'.format(dataset))

        try:
            from rpy2.robjects.packages import importr, data
        except ImportError: # pragma: no cover
            raise ImportError(
                "This example needs Rpy2."
                "Please refer to the R requirements in the README"
            )        
        datasets = importr('datasets')
        r_env = data(datasets).fetch(dataset)

        import rpy2.robjects as robjects
        # why we do this:
        #> data(euro)
        #> euro
        #ATS         BEF         DEM         ESP         FIM         FRF         IEP         ITL         LUF         NLG         PTE 
        #13.760300   40.339900    1.955830  166.386000    5.945730    6.559570    0.787564 1936.270000   40.339900    2.203710  200.482000 
        #
        #> as.data.frame(euro)
        #        euro
        #ATS   13.760300
        #BEF   40.339900
        #DEM    1.955830
        data = robjects.r('as.data.frame(%s)' % dataset)

        # at time of writing, rpy2's R dataframe to pandas dataframe was not fully supported
        # However, as python list() seems to work for FloatVector, StrVector, and FactorVector, let's use it
        from rpy2.robjects import r
        colnames = r.colnames(data)
        pandas_data = {}
        # convert each column of the R dataframe in turn
        for i, colname in enumerate(colnames):
            pandas_data[colname] = list(data[i])
        # Unfortunately, some datasets have rownames that should be an ID column (e.g., see mtcars where rownames=names of the cars). 
        # This is the best we can do: pull it out as an additional column for each and every dataset
        pandas_data['row_names'] = list(data.rownames)

        df = pd.DataFrame(pandas_data)
        data_object.add(self, df)
        terminate = df.empty
        return data_object, terminate



import click
import logging
import os
import sys
import importlib
import pkg_resources

logging.basicConfig(format='%(asctime)s %(levelname)s %(filename)s %(funcName)s: %(message)s', level=logging.INFO)


def replace_line(file_name, line_num, text):
    """Replace a single line of a file with text

    Args:
        file_name (str): name of file
        line_num (int): line to replace
        text (str): string to replace at line_num

    Returns:
        Nothing, the file is modified and rewritten

    """
    lines = open(file_name, 'r').readlines()
    lines[line_num] = text
    out = open(file_name, 'w')
    out.writelines(lines)
    out.close()


@click.group()
def cli():
    """This is the command line interface for primrose.
    The command most people need is `primrose run` to run a primrose job.
    Type primrose commandname --help for more detailed help on a command"""
    pass


@click.command()
@click.option('--config', required=True, help="Path to Config file")
def validate(config):
    """Validate a primrose config"""
    from primrose.configuration.configuration import Configuration
    configuration = Configuration(config_location=config)


@click.command()
@click.option('--config', required=True, help="Path to Config file")
@click.option('--dry_run', default=False, help="Config file")
def run(config, dry_run=False):
    """Run a primrose job"""
    from primrose.configuration.configuration import Configuration
    from primrose.dag_runner import DagRunner

    configuration = Configuration(config_location=config)
    DagRunner(configuration).run(dry_run=dry_run)


@click.command()
@click.option('--config', required=True, help="Path to Config file")
@click.option('--node_size', help='Size of nodes', required=False, default=500)
@click.option('--label_font_size', help='Size of font for text labels', required=False, default=12)
@click.option('--text_angle', help='Angle to rotate text (counterclockwise)', required=False, default=0)
@click.option('--image_width', help='width of image in inches', required=False, default=16)
@click.option('--image_height', help='height of image in inches', required=False, default=12)
@click.option('--nodesequence', help='Show node sequences numbers? "true" or "false"?', required=False)
@click.option('--outfile', help='Path for the output image file', required=True)
def plot(config, node_size, label_font_size, text_angle, image_width, image_height, nodesequence, outfile):
    '''Create an image of the DAG'''
    from primrose.configuration.configuration import Configuration
    from primrose.dag.traverser_factory import TraverserFactory

    config = Configuration(config_location=config)

    filename = outfile
    other_args = {}
    other_args['filename'] = outfile

    if node_size:
        other_args['node_size'] = int(node_size)

    if label_font_size:
        other_args['label_font_size'] = int(label_font_size)

    if text_angle:
        other_args['text_angle'] = int(text_angle)

    if image_width:
        other_args['image_width'] = int(image_width)

    if image_height:
        other_args['image_height'] = int(image_height)

    other_args['traverser'] = None
    if nodesequence and nodesequence.lower() == "true":
        traverser = TraverserFactory().default_traverser(config)

        if config.config_metadata and 'traverser' in config.config_metadata:
            traverser = TraverserFactory().instantiate(config.config_metadata['traverser'], config)
            logging.info("Setting to Traverser to %s", config.config_metadata['traverser'])

        other_args['traverser'] = traverser

    logging.info("passing in %s", other_args)
    config.dag.plot_dag(**other_args)


@click.command()
@click.option('--destination', required=True, help="Path to destination")
def generate_run_script(destination):
    """Create primrose run script in your project--if you need to register your own primrose classes.
    This will also allow you to register and use your own primrose node classes. 
    To register your own classes you will also need to run the primrose generate_class_registration_template command"""
    from shutil import copyfile
    run_primrose_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates', 'run_primrose.py')
    copyfile(run_primrose_path, destination)
    print("Primrose's run_primrose.py script copied to " + destination)
    print("Don't forget to set where your new classes are to be imported. See script.")


@click.command()
@click.option('--destination', required=True, help="Path to destination")
def generate_class_registration_template(destination):
    """Create template to register your own classes. This will copy a template to your destination, such as src/__init__.py"""
    from shutil import copyfile
    user_registration_template_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                                   'templates',
                                                   'user_registration_template.py')
    copyfile(user_registration_template_path, destination)
    print("Primrose's user_registration_template.py script copied to " + destination)

@click.command()
@click.option('--name', required=True, help="Name of primrose project to be generated")
def create_project(name):
    """Create template project filled with example configuration"""

    from shutil import copyfile, copytree

    # get local directories to copy from the package dir
    package_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')
    env_dir = sys.prefix

    config_dir_path = os.path.join(env_dir, 'config')

    data_dir_path = os.path.join(env_dir, 'data')

    user_registration_template_path = os.path.join(package_dir, 'user_registration_template.py')

    run_primrose_path = os.path.join(package_dir, 'run_primrose.py')

    # make empty cache and src dir
    os.makedirs(os.path.join(name, 'cache'))
    os.makedirs(os.path.join(name, 'src', 'yourpackage'))

    # copy relevant files to user path
    copytree(config_dir_path, os.path.join(name, 'config'))

    copytree(data_dir_path, os.path.join(name, 'data'))

    copyfile(os.path.join(package_dir, 'awesome_model.py'), os.path.join(name,
                                                                         'src',
                                                                         'yourpackage',
                                                                         'awesome_model.py'))

    copyfile(os.path.join(package_dir, 'awesome_reader.py'), os.path.join(name,
                                                                          'src',
                                                                          'yourpackage',
                                                                          'awesome_reader.py'))

    copyfile(user_registration_template_path, os.path.join(name, 'src', '__init__.py'))

    copyfile(run_primrose_path, os.path.join(name, 'run_primrose.py'))

    # modify run_primrose to take into account the templated extention functions
    replacement_line = 'from src.__init__ import *\n'

    replace_line(os.path.join(name, 'run_primrose.py'), 14, replacement_line)

    # make __init__ file for the yourpackage directory
    with open(os.path.join(name, 'src', 'yourpackage','__init__.py'), 'w') as f:
        f.write('')

    print("New primrose project, {} built!".format(name))

@cli.command()
def version():
    '''Print the installed primrose version'''
    print(pkg_resources.get_distribution("primrose").version)

cli.add_command(validate)
cli.add_command(run)
cli.add_command(plot)
cli.add_command(generate_run_script, name='generate-run-script')
cli.add_command(generate_class_registration_template, name='generate-class-registration-template')
cli.add_command(create_project, name='create-project')
cli.add_command(version)

if __name__ == "__main__":
    cli()



"""Module with abstract pipeline class to specify interface needed for future pipelines

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from abc import abstractmethod
from primrose.base.node import AbstractNode

class AbstractConditionalPath(AbstractNode):
    """A class that supports conditional pathing through the DAG. 
    After running, prune() can provide a list of destination nodes, representing start of paths to prune. 
    DAGRunner can then prune those nodes, and all paths downstream of those nodes, from the DAG"""

    def all_nodes_to_prune(self):
        """What are all the nodes we should prune from the DAG?

        Note:
            this call destinations_to_prune() and then uses the DAG to identify 
            the complete subgraphs starting with those destinations

        Returns:
            set of all nodes to prune

        """
        if not 'destinations' in self.node_config:
            raise Exception("Destinations key is missing")

        destinations_to_prune = self.destinations_to_prune()

        if destinations_to_prune:

            all_nodes_to_prune = set()

            for destination in destinations_to_prune:
                if not destination in self.node_config['destinations']:
                    raise Exception("Destination " + destination + " is not in destinations list")

                all_nodes_to_prune.add(destination)
                all_nodes_to_prune.update(self.configuration.dag.descendents(destination))

            return all_nodes_to_prune

        return None

    @abstractmethod
    def destinations_to_prune(self):
        """Which destinations, if any, should we prune from DAG?

        Returns:
            list of destinations nodes to prune from DAG, None otherwise

        """
        return None # pragma: no cover


"""Helper function for get os.environ values

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os

def get_env_val(key):
    """get environmental variable

    Returns:
        env variable (str)

    """
    val = os.environ.get(key)
    if val is None:
        raise Exception("Did not find env variable for " + str(key))
    return val




"""Abstract reader that uses SQL

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os
from primrose.base.reader import AbstractReader
from abc import abstractmethod
import pandas as pd
import logging
import hashlib
from jinja2 import Environment, FileSystemLoader

class AbstractSqlReader(AbstractReader):
    """A reader that explicitly reads from relational DB using SQL and is able to run pd.read_sql."""

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            After adding this list, validation automatically occurs before instantiation in the pipeline factory.

        Returns:
            set of keys necessary to run implementation

        """
        return set(['query_json']) # pragma: no cover

    @staticmethod
    def _substitute_query(query_json):
        """Substitue our paramters in a query string. Renders the file as a jinja template to allow for advanced
        query nesting and logic. Uses a jinja environment which searches for templates (sql queries) using relative 
        path and absolute path.
        
        Note:
            given JSON:

            {
                "query": "somequery.sql",
                "parameters": {
                    "x": 3,
                    "y": "somestring"
                }
            }

            and `somequery.sql`:
            ```
            SELECT * from table_01 where x = {x} and y = {y}
            ```

            This function will read in the SQL file via jinja, and if parameters is present, substitue out the parameters, returning the final query string

        Returns:
            query_json (JSON): JSON

        """
        assert 'query' in query_json
        assert os.path.exists(query_json['query'])

        sql_input = {}
        if 'parameters' in query_json:
            sql_input = query_json['parameters']

        jinja_env = Environment(loader=FileSystemLoader(['.', '/']), 
                                variable_start_string='{',
                                variable_end_string='}')
        
        query = jinja_env.get_template(query_json['query'])
        return query.render(**sql_input)

    def _generate_queries(self):
        """generate final queries, using parameter substitution, if requested

        Note:
            query_json should be structured as: \
                "query_json": [ \
                    { \
                        "query": "somequery.sql", \
                        "parameters": { \
                            "x": 1, \
                            "y": 2 \
                        } \
                    }, \
                    { \
                        "query": "someotherquery.sql", \
                        "parameters": { \
                            "x": 3, \
                            "y": "somestring" \
                        } \
                    } \
                ] \
            where parameters is an optional key.

            This generator will return each query string with variables substituted from parameters (if any) 

        Yields:
            query (str): query

        """
        for individual_query_json in self.node_config['query_json']:
            yield self._substitute_query(individual_query_json)

    def _get_key(self, i):
        """Returns the key name for the ith query in the query_json list. This is the key used when adding the dataframe to the dataobject.
        If none is provided, default to `query_i`

        Args:
            i (int): the index of the query
        
        Returns:
            key (str): key name for placement in a `DataObject`
        """
        key = "query_" + str(i)
        this_query_config = self.node_config['query_json'][i]
        if "key" in this_query_config.keys():
            key = this_query_config['key']
        return key

    @abstractmethod
    def get_connection(self):
        """return a database connection, one that is compatible with pd.read_sql"""
        pass # pragma: no cover

    def run(self, data_object):
        """run SQL queries into pandas dataframes

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        conn = self.get_connection()
        debug = self.node_config.setdefault('debug',False)
        for i, query in enumerate(self._generate_queries()):

            if debug:
                with open('debug_query_' + str(i) + '.sql', 'w') as qfile:
                    qfile.write(os.path.join(query))
            df = self.query_db(query,conn)
            key = self._get_key(i)
            logging.info("Adding df with key %s", key)
            data_object.add(self, df, key)
            if df.empty:
                return data_object, True

        terminate = False
        return data_object, terminate

    def query_db(self, query, conn):
        """Query the db using Bigquery logic if specified.

        Args:
            query (str): SQL query string
            conn (DB connection): database connection

        Returns:
            dataframe (DataFrame): dataframe

        """
        return pd.read_sql(query, con=conn)



"""Abstract class to write some data to a some file

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import pathlib
from primrose.base.writer import AbstractWriter

class AbstractFileWriter(AbstractWriter):
    """write some data to a some file"""

    def __init__(self,configuration, instance_name):
        """

        Args:
            configuration (Configuration): configuration object defined in primrose/Configuration with validated inputs
                from the result of necessary_config, all inputs are described in that method

            instance_name (str): how the code knows where it is from

        """
        super().__init__(configuration, instance_name)
        self._check_directory()

    def _check_directory(self):
        """Create directory or directory structure if it does not already exist."""
        pathlib.Path(self.node_config['dir']).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def necessary_config(node_config):
        """Necessary inputs for file writers:

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            key: key that identifies object to write
            dir: directory to write csv (must already exist)
            filename: name of file to be written

        Returns:
            set of necessary configuration keys

        """
        return set(["key", "dir", "filename"])



import pytest
from primrose.readers.sklearn_dataset_reader import SklearnDatasetReader
from primrose.configuration.configuration import Configuration
from primrose.data_object import DataObject, DataObjectResponseType

def test_run():
    config = {
        "implementation_config":{
            "reader_config": {
                "dataset_reader": {
                    "class": "SklearnDatasetReader",
                    "dataset": "iris",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)
    reader = SklearnDatasetReader(configuration, "dataset_reader")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    df = data_object.get("dataset_reader", rtype=DataObjectResponseType.VALUE.value)
    assert len(df.columns) == 5
    assert "target" in df.columns
    assert df.shape == (150, 5)
    assert 'sepal length (cm)' in df.columns


"""concrete class that performs TFIDF on lemmatized tokens with optional ngrams

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from nltk import ngrams
from primrose.base.search_engine import AbstractSearchEngine
from nltk import WordNetLemmatizer

class MinimalSearchEngine(AbstractSearchEngine):
    """simple TFIDF search engine"""

    def __init__(self, configuration, instance_name):
        """instantiate the search engine

        Args:
            configuration (Configuration): Configuration instance
            instance_name (str): name of instance

        """
        AbstractSearchEngine.__init__(self, configuration, instance_name)
        self.lemmatizer = WordNetLemmatizer()

    def tokenize(self, s, stopwords=[], add_ngrams=True):
        """ tokenize a string document, optimized for recipe names given default stopwords and other
            string cleanup operations

        Args:
            s (str): some document string
            stopwords (list): list of stopwords
            add_ngrams (bool): whehter to include ngrams on tokens

        Returns
            tokens (list): list of cleaned, standardized tokens from document

        """
        q = s.lower().strip()
        tokens = q.replace.split(" ")
        tokens = [w for w in tokens if w not in stopwords]
        tokens = [self.lemmatizer.lemmatize(w) for w in tokens]
        if add_ngrams:
            bigrams = list(ngrams(tokens, 2))
            strbigrams = ["_".join(t) for t in bigrams]
            tokens.extend(strbigrams)
        return tokens



import logging
import argparse

from primrose.configuration.configuration import Configuration
from primrose.dag.traverser_factory import TraverserFactory

def parse_arguments():
    """Parse command line arguments
    Use environment variables as default if passed.
    Returns: argument objects with flags as attributes
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--config_loc',
                        help='Location of the configuration file',
                        required=True)
    parser.add_argument('--node_size',
                        help='Size of nodes',
                        required=False)
    parser.add_argument('--label_font_size',
                        help='Size of font for text labels',
                        required=False)
    parser.add_argument('--text_angle',
                        help='Angle to rotate text (counterclockwise)',
                        required=False)
    parser.add_argument('--image_width',
                        help='width of image in inches',
                        required=False)
    parser.add_argument('--image_height',
                        help='height of image in inches',
                        required=False)
    parser.add_argument('--nodesequence',
                        help='Show node sequences numbers? "true" or "false"?',
                        required=False)
    parser.add_argument('--outfile',
                        help='Path for the output image file',
                        required=True)
    known_args, pipeline_args = parser.parse_known_args()
    return known_args, pipeline_args

def main():
    """Train, evaluate or predict with a machine learning model from a user defined data source
        The function can optionally upload the results to an external source
    """
    args, _ = parse_arguments()

    logging.basicConfig(format='%(asctime)s %(filename)s %(funcName)s: %(message)s', level=logging.INFO)
    logging.getLogger().setLevel(logging.INFO)

    config = Configuration(config_location=args.config_loc)

    filename = args.outfile
    other_args = {}
    other_args['filename'] = args.outfile

    if args.node_size:
        other_args['node_size'] = int(args.node_size)

    if args.label_font_size:
        other_args['label_font_size'] = int(args.label_font_size)

    if args.text_angle:
        other_args['text_angle'] = int(args.text_angle)

    if args.image_width:
        other_args['image_width'] = int(args.image_width)

    if args.image_height:
        other_args['image_height'] = int(args.image_height)

    other_args['traverser'] = None
    if args.nodesequence and args.nodesequence.lower() == "true":
        traverser = TraverserFactory().default_traverser(config)

        if config.config_metadata and 'traverser' in config.config_metadata:
            traverser = TraverserFactory().instantiate(config.config_metadata['traverser'], config)
            logging.info("Setting to Traverser to %s", config.config_metadata['traverser'])

        other_args['traverser'] = traverser

    logging.info("passing in %s", other_args)
    config.dag.plot_dag(**other_args)

if __name__ == '__main__':
    main()



"""Singleton Factory where one can register objects/classes for instantiation

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""

import logging
import inspect
import sys
from primrose.base.node import AbstractNode

from primrose.readers.csv_reader import CsvReader
from primrose.readers.gcs_dill_reader import GcsDillReader
from primrose.readers.dill_reader import DillReader
from primrose.readers.deserializer import Deserializer, GcsDeserializer
from primrose.readers.sklearn_dataset_reader import SklearnDatasetReader
from primrose.readers.mysql_reader import MySQLReader
from primrose.readers.sqlite_reader import SQLiteReader
from primrose.readers.postgres_reader import PostgresReader

from primrose.conditionalpath.simple_switch import SimpleSwitch

from primrose.pipelines.dataframe_joiner import DataFrameJoiner
from primrose.pipelines.encode_train_test_split import EncodeTrainTestSplit
from primrose.pipelines.train_test_split import TrainTestSplit
from primrose.pipelines.sklearn_preprocessing_pipeline import SklearnPreprocessingPipeline

from primrose.models.sklearn_classifier_model import SklearnClassifierModel
from primrose.models.sklearn_cluster_model import SklearnClusterModel
from primrose.models.sklearn_regression_model import SklearnRegressionModel

from primrose.notifications.success_notification import ClientNotification

from primrose.writers.csv_writer import CsvWriter
from primrose.writers.file_writer import FileWriter
from primrose.writers.dill_writer import DillWriter
from primrose.writers.serializer import Serializer
from primrose.writers.s3_writer import S3Writer

from primrose.dataviz.cluster_plotter import ClusterPlotter

from primrose.cleanup.logging_success import LoggingSuccess
from primrose.readers.r_reader import RReader

class NodeFactory:
    """Singleton Factory where one can register objects/classes for instantiation"""
    #https://python-3-patterns-idioms-test.readthedocs.io/en/latest/Singleton.html

    instance = None

    CLASS_KEY = "class"
    CLASS_PREFIX = "class_prefix"

    def __init__(self):
        """instantiate the factory but as a singleton. The guard raails are here
        """
        # where the magic happens, only one instance allowed:
        if not NodeFactory.instance:
            NodeFactory.instance = NodeFactory.__HiddenFactory()

    def __getattr__(self, name):
        """getattr with instance name

        Args:
            name (str): name of the instance

        Returns:
            gettattr

        """
        return getattr(self.instance, name)

    class __HiddenFactory:
        """actual factory where registry and instantiation happens"""

        def __init__(self):
            """instantiate the HiddenFactory"""
            self.name_dict = {'CsvReader': CsvReader,
                'SQLiteReader': SQLiteReader,
                'MySQLReader': MySQLReader,
                'PostgresReader': PostgresReader,
                'GcsDillReader': GcsDillReader,
                'DillReader': DillReader,
                'Deserializer': Deserializer,
                'GcsDeserializer': GcsDeserializer,
                'DataFrameJoiner': DataFrameJoiner,
                'EncodeTrainTestSplit': EncodeTrainTestSplit,
                'TrainTestSplit': TrainTestSplit,
                'CsvWriter': CsvWriter,
                'FileWriter': FileWriter,
                'DillWriter': DillWriter,
                'Serializer': Serializer,
                'S3Writer': S3Writer,
                'SklearnClassifierModel': SklearnClassifierModel,
                'LoggingSuccess': LoggingSuccess,
                'ClusterPlotter': ClusterPlotter,
                'SklearnClusterModel': SklearnClusterModel,
                'SklearnPreprocessingPipeline': SklearnPreprocessingPipeline,
                'SklearnDatasetReader': SklearnDatasetReader,
                'SklearnRegressionModel': SklearnRegressionModel,
                'SimpleSwitch': SimpleSwitch,
                'ClientNotification': ClientNotification,
                'RReader': RReader}

        def register(self, key, class_obj, raise_on_overwrite=False):
            """Registering class_obj with key

            Args:
                key (str): key such as class name, e.g. 'BQWriter'
                class_obj (class obj), e.g. BQWriter

            Returns:
                nothing. Side effect is to register the class

            """
            if raise_on_overwrite and key in self.name_dict:
                raise Exception("Node already exist with the key " + key)

            if not (inspect.isclass(class_obj) and issubclass(class_obj, AbstractNode)):
                raise Exception("NodeFactory can only register classes that implement AbstractNode")

            if key is None:
                try:
                    key = class_obj.__name__
                except AttributeError as e:
                    raise Exception(f"Cannot register {class_obj}, no __name__ attribute found. Please explicity specify a name when registering this class.")

            self.name_dict[key] = class_obj
            logging.debug("Registered %s : %s" % (key, class_obj))

        def is_registered(self, class_key):
            """is this class registered?

            Args:
                class_key (str): key used to register class

            Returns:
                determination (boolean) of whether this is already register

            """
            return class_key in self.name_dict

        def unregister(self, key):
            """unregister an entry

            Arguments:
                key (str): key to unregister

            Returns:
                nothing. Side effect is that the object is unregistered

            """
            if key in self.name_dict:
                del self.name_dict[key]
                logging.info("Unregistered %s", key)
            else:
                raise Exception("Key not found %s" % key)

        def instantiate(self, class_name, configuration, instance_name):
            '''instantiate instances of node, given name key of node (typically a classname but is not required to be)

            Args:
                class_name (str): name key for the class
                configuration (Configuration): Configuration instance

            Returns:
                instance (Traverser): instance of a traverser

            '''
            return self.name_dict[class_name](configuration, instance_name)

        def valid_configuration(self, instance_class, configuration_dict):
            """Check for all the correct configuration keys via the necessary_config method

                For instance, if a CSVWriter has a required filename key,
                check that it exists as a key in the configuration

            Args:
                instance_class: class to validate
                configuration_dict: configuration dictionary from a Configuration object

            Returns:
                True for valid configuration

            Raises:
                Exception if keys not valid

            """
            configuration_requirements = instance_class.necessary_config(configuration_dict)

            if all([key in configuration_dict for key in configuration_requirements]):
                return True
            else:
                raise Exception('Configuration missing necessary keys for '
                                '{}, required keys: {}\n\n{}'.format(instance_class,
                                                                    configuration_requirements,
                                                                    instance_class.necessary_config.__doc__))

        def register_module_classes(self, module):
            """Given some module, find and register any AbstractNode classes it finds in that module

            Args:
                module (str): e.g. __name__, 'src'

            Returns:
                nothing. Side effect is to register the Node classes found

            """
            for name, obj in inspect.getmembers(sys.modules[module]):
                if inspect.isclass(obj) and issubclass(obj, AbstractNode):
                    logging.info("Discovered class " + name + " (" + str(obj) + ")")
                    self.register(name, obj)


"""Simple conditional path node that will travel down one destination and prune the rest

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.conditional_path_node import AbstractConditionalPath

class SimpleSwitch(AbstractConditionalPath):

    @staticmethod
    def necessary_config(node_config): 
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        Returns:
            set of keys necessary to run implementation

        """
        return set(['path_to_travel'])

    def destinations_to_prune(self):
        """List destinations that the node wishes to be pruned from the DAG
        
        Returns:
            destinations (list): list of destinations

        """
        destinations = self.node_config['destinations']
        destinations.remove(self.node_config['path_to_travel'])
        return destinations

    def run(self, data_object):
        """Run this node. Here, do nothing

        Args:
            data_object (DataObject): DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        return data_object, False



import pytest
from primrose.cleanup.logging_success import LoggingSuccess
from primrose.configuration.configuration import Configuration
from testfixtures import LogCapture
from primrose.data_object import DataObject

def test_necessary_config():
    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['successlogger']
                }
            },
            "cleanup_config": {
                "successlogger": {
                    "class": "LoggingSuccess",
                    "msg": "woohoo, all done!",
                    "level": "INFO",
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    ls = LoggingSuccess(configuration, "successlogger")
    node_config = {
                    "class": "LoggingSuccess",
                    "msg": "woohoo, all done!",
                    "level": "INFO",
                }
    assert isinstance(ls.necessary_config(node_config), set)
    assert len(ls.necessary_config(node_config)) == 2

    data_object = DataObject(configuration)

    assert ls.run(data_object)

    with LogCapture() as l:
        ls.run(data_object) 
    l.check(('root', 'INFO', 'woohoo, all done!'),)



'''
    Run a job: i.e. run a configuration file through the DAGRunner
'''
import argparse
import logging
import warnings

######################################
######################################
# Important:
#
# If your configuration uses custom node classes, be sure to set environment variable
# PRIMROSE_EXT_NODE_PACKAGE to the location of your package before running primrose.
# Example:
#   ```
#   export PRIMROSE_EXT_NODE_PACKAGE=src/mypackage
#   python run_primrose.py --config_loc my_config.json
#   ```
#
######################################
######################################

from primrose.configuration.configuration import Configuration
from primrose.dag_runner import DagRunner
from primrose.dag.config_layer_traverser import ConfigLayerTraverser
from primrose.dag.depth_first_traverser import DepthFirstTraverser

warnings.filterwarnings("ignore")

def parse_arguments():
    """
        Parse command line arguments

        Returns:
            argument objects with flags as attributes
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--config_loc',
                        help='Location of the configuration file',
                        required=True)
    parser.add_argument('--is_dry_run',
                        help='do a dry run of the DAG which will validatre config and log which nodes would be run',
                        default=False,
                        type=lambda x: (str(x).lower() == 'true'))

    known_args, pipeline_args = parser.parse_known_args()
    return known_args, pipeline_args

def main():
    """
        Run a job: i.e. run a configuration file through the DAGRunner
    """
    args, _ = parse_arguments()

    logging.basicConfig(format='%(asctime)s %(levelname)s %(filename)s %(funcName)s: %(message)s', level=logging.INFO)

    configuration = Configuration(config_location=args.config_loc)

    DagRunner(configuration).run(dry_run=args.is_dry_run)

if __name__ == '__main__':
    main()



"""write some data to a S3 bucket, via a local file.

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os
import uuid
import logging
import boto3

from primrose.base.writer import AbstractWriter
from primrose.data_object import DataObjectResponseType


class S3Writer(AbstractWriter):
    """a writer that writes a dataframe to disk and from there to S3 bucket"""

    @staticmethod
    def necessary_config(node_config):
        """Necessary keys for S3Writer object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            dir: directory to write, then upload file
            bucket_name: s3 bucket where the file will be written
            bucket_filename: filename for the s3 bucket

        Returns:
            set of necessary config fields

        """
        return set(["dir", "bucket_name", "bucket_filename"])

    def _write_locally(self, data_object):
        """write data to local file

        Args:
            data_object (DataObject): instance of DataObject

        Note:
            generates unique filename using uuid

        Returns:
            filename (str)

        """
        # save dataframe to CSV file first

        unique_filename = str(uuid.uuid4())
        filename = self.node_config['dir'] + os.path.sep + unique_filename
        key = self.node_config['key']
        logging.info("Saving %s data to %s", key, filename)

        data_to_write = data_object.get_upstream_data(self.instance_name, pop_data=False, rtype=DataObjectResponseType.KEY_VALUE.value)
        assert key in data_to_write
        data_to_write[key].to_csv(filename, index=False)

        assert os.path.exists(filename)
        return  filename

    def run(self, data_object):
        """write some data to a local CSV file then from there to S3 bucket

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        filename = self._write_locally(data_object)

        # Create an S3 client
        # Boto3 will check these environment variables for credentials:
        # AWS_ACCESS_KEY_ID
        # AWS_SECRET_ACCESS_KEY
        s3 = boto3.client('s3')

        # Uploads the given file using a managed uploader, which will split up large
        # files automatically and upload parts in parallel.
        bucket_name = self.node_config['bucket_name']
        key = self.node_config['bucket_filename']
        logging.info("uploading to %s %s", bucket_name, key)
        s3.upload_file(filename, bucket_name, key)

        # TODO check for success
        # see https://stackoverflow.com/questions/41565766/make-a-unittest-for-a-function-that-uploads-a-file-to-s3-using-boto3
        #s3 = boto3.resource('s3')
        #mytname = nameOfFile
        #obj = s3.Object(bucket_name='cloud-test-cf', key=mytname)
        #response = obj.get()
        #self.assertEqual(response['ContentLength'], len(contentsOfFile))
        #remoteData = response['Body'].read()
        #self.assertEqual(remoteData, contentsOfFile)

        logging.info("upload complete!")
        terminate = False
        return data_object, terminate


"""Module to implement a Configuration parser which enhances parsing functionality of configparser

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import re
import datetime
import jstyleson
import yaml
import json
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
import hashlib
import os
import logging
import importlib
import glob
from primrose.node_factory import NodeFactory
from primrose.configuration.util import OperationType, ConfigurationError, ConfigurationSectionType
from primrose.configuration.configuration_dag import ConfigurationDag
from primrose.dag.traverser_factory import TraverserFactory


SUPPORTED_EXTS = frozenset(['.json', '.yaml', '.yml'])
CLASS_ENV_PACKAGE_KEY = 'PRIMROSE_EXT_NODE_PACKAGE'

class Configuration:
    """Stores user defined configuration for primrose job"""

    def __init__(self, config_location, is_dict_config=False, dict_config=None):
        """Read in configuration file and parse into specified values

        Args:
            config_location (str): valid filepath for file
            is_dict_config (bool): are we passing in a dictionary configuration directly
            dict_config (dict): dictionary object, if is_dict_config

        """
        if is_dict_config:
            ext = None

            if dict_config is None:
                raise Exception('expected dict_config was None')

            if not isinstance(dict_config, dict):
                raise Exception('did not receive expected dict_config')

            dict_str = jstyleson.dumps(dict_config)

            config_str = Configuration.perform_any_config_fragment_substitution(dict_str)

        else:
            logging.info('Loading config file at {}'.format(config_location))
            self.config_location = config_location

            if os.path.exists(config_location):
                ext = os.path.splitext(config_location)[1].lower()
                if ext not in SUPPORTED_EXTS:
                    raise ValueError('config file at: {} has improper extension type - please use a .json or .yml file'.format(config_location))

                with open(config_location, 'r') as f:
                    config_str = f.read()

                config_str = Configuration.perform_any_config_fragment_substitution(config_str)

            else:
                raise Exception('config file at: {} not found'.format(config_location))

        if ext is None or ext == '.json':
            self.config = jstyleson.loads(config_str, object_pairs_hook=self.dict_raise_on_duplicates)
        elif ext in ['.yaml', '.yml']:
            self.config = yaml.load(config_str, Loader=yaml.FullLoader)

        assert isinstance(self.config, dict)

        # check top-level keys
        for k in self.config:
            if k not in ConfigurationSectionType.values():
                msg = "Unsupported top-level key: %s. " % k
                msg += "Supported keys are %s" % str(ConfigurationSectionType.values())
                raise ConfigurationError(msg)

        # metadata section can be optional
        self.config_metadata = None
        if ConfigurationSectionType.METADATA.value in self.config:
            self.config_metadata = self.config[ConfigurationSectionType.METADATA.value]

        # implemetation_config section is required
        if not ConfigurationSectionType.IMPLEMENTATION_CONFIG.value in self.config:
            raise ConfigurationError("Did not find required top-level key %s" % ConfigurationSectionType.IMPLEMENTATION_CONFIG.value)

        # keep a copy of the complete configuration
        self.complete_config = self.config.copy()

        # note: config is now just the implementation component of the dictionary
        self.config = self.config[ConfigurationSectionType.IMPLEMENTATION_CONFIG.value]

        # store the dag object
        self.dag = ConfigurationDag(self.config)

        # populate configuration file string and hash
        self.config_string, self.config_hash = self._get_configuration_hash()

        # get the formatted time this file was instantiated
        self.config_time = datetime.datetime.now().strftime('%Y%m%d_%H%M')

        # parse the file into an internal config object
        self._parse_config()

        self.check_config()

    @staticmethod
    def perform_any_config_fragment_substitution(config_str):
        """Given some configuration file content string, look for \
        subtitutions given by `$$FILE=path/to/config/file/fragment.json$$` and make the \
        replacements using the filenames provided\
        For example: \
        { \
            $$FILE=/tmp/metadata.json$$ \
            "implementation_config": { \
                $$FILE= config/read_write_fragment.json $$ \
            } \
        } \
        will inject /tmp/metadata.json into the 2nd line of that config.

        Args:
            config_str (str): content of some configuration file that may or may not contain substition variables

        Returns:
            config_str (str): the post-substituted configuration string

        """
        jinja_env = Environment(loader=FileSystemLoader(['.', '/']))
        try:
            config_str_template = jinja_env.from_string(config_str)
            config_str = config_str_template.render()
        except(TemplateNotFound) as error:
            filenames = str(error)
            raise ConfigurationError(f"Substitution files do not exist: {filenames}")
        return config_str

    def dict_raise_on_duplicates(self, ordered_pairs):
        """Reject duplicate keys in JSON string, ie. sections and node names.

        Args:
            ordered_pairs (list): list of key:values from the config \
            Example: ordered_pairs `[('class', 'CsvReader'), ('filename', 'data/tennis.csv'), ('destinations', ['write_output'])]` \
            ordered_pairs `[('read_data', {'class': 'CsvReader', 'filename': 'data/tennis.csv', 'destinations': ['write_output']})]`

        Returns:
            dictionary (dict): dictionary of key (node type) and value (node name)

        """
        #https://stackoverflow.com/questions/14902299/json-loads-allows-duplicate-keys-in-a-dictionary-overwriting-the-first-value

        d = {}
        for k, v in ordered_pairs:
            if k in d:
                raise ConfigurationError("duplicate key: %r" % (k,))
            else:
                d[k] = v
        return d

    def _get_configuration_hash(self):
        """Get configuration file string and hash

        Returns:
            (tuple): tuple containing:

                configuration_string (str):  configuration_string

                configuration_file_hashname (str): terminate the DAG?

        """
        configuration_string = json.dumps(self.complete_config, sort_keys=True)
        configuration_file_hashname = hashlib.sha256(configuration_string.encode('utf-8')).hexdigest()
        return configuration_string, configuration_file_hashname

    def check_metadata(self):
        """checks some dependencies among metadata keys

        Raises:
            ConfigurationError is issues found

        """
        if self.config_metadata:

            if 'traverser' in self.config_metadata:
                classname = self.config_metadata['traverser']
                try:
                    TraverserFactory().instantiate(classname, self)
                except KeyError:
                    raise Exception(classname + " is not a valid and/or registered Traverser")

            if 'data_object' in self.config_metadata:

                cfg = self.config_metadata['data_object']

                if 'read_from_cache' in cfg and (cfg['read_from_cache'] or str(cfg['read_from_cache']).lower() == "true"):

                    if not 'read_filename' in cfg:
                        raise ConfigurationError("metadata.data_object: if read_from_cache==true, you must set 'read_filename'")

                    #just check path exists but not that one can read into a DataObject
                    if not os.path.exists(cfg['read_filename']):
                        raise ConfigurationError("Invalid metadata.data_object.read_filename: " + str(cfg['read_filename']))

                if 'write_to_cache' in cfg and (cfg['write_to_cache'] or str(cfg['write_to_cache']).lower() == "true"):

                    if not 'write_filename' in cfg:
                        raise ConfigurationError("metadata.data_object: if write_to_cache==true, you must set 'write_filename'")

    def check_sections(self):
        """Check that all the sections in implementation are supported ones.
        Either the user supplied metata.section_registry, or they are using default sections

        Raises:
            ConfigurationError if declaring metadata.section_registry and sections from implementation were not found in metadata
            or vice versa, or if using default operations but sections found that were not supported

        """
        if self.config_metadata and 'section_registry' in self.config_metadata and len(self.config_metadata['section_registry']) > 0:
            actual_set = set(self.config.keys())
            user_set = set(self.config_metadata['section_registry'])

            if actual_set != user_set:

                diff = user_set.difference(actual_set)
                if len(diff) > 0:
                    msg = "Following sections from metadata were not found implementation: " + str(diff)
                    raise ConfigurationError(msg)

                diff = actual_set.difference(user_set)
                if len(diff) > 0:
                    msg = "Following sections from implementation were not found in metadata: " + str(diff)
                    raise ConfigurationError(msg)

            logging.info("OK: section_registry sections match implementation sections")
            return

        # otherwise, let's check for default operations
        supported = OperationType.values()
        for section_key in self.config.keys():
            if section_key not in supported:
                raise ConfigurationError("Unspported operation: %s" % section_key)
        logging.info("OK: all sections are supported operations")

    def config_for_instance(self, instance_name):
        """get the configuration for a given node / instance_name

        Returns:
            JSON chunk for this instance

        """
        if not instance_name in self.instance_to_config:
            raise Exception("Unknown key " + instance_name)
        return self.instance_to_config[instance_name]

    def sections_in_order(self):
        """Return list of section names in order, either explicitly from metadata or from default Enum order

        Note:
            If there is a non-empty section_run list in metadata return that
            elif there is a non-empty section_registry in metadata return that
            otherwise return sections present from default OperationType enum.

            We need this method because the config sections are a dictionary not a list so we can't
            guarantee order of keys. This method imposes an expected order.

        Returns:
            (tuple): tuple containing:

                section names (list): list of sections

                source (str): where did the list come from? section_run, section_registry, or default?

        """
        if self.config_metadata and 'section_run' in self.config_metadata and len(self.config_metadata['section_run']) > 0:
            return self.config_metadata['section_run'], 'section_run'

        if self.config_metadata and 'section_registry' in self.config_metadata and len(self.config_metadata['section_registry']) > 0:
            return self.config_metadata['section_registry'], 'section_registry'

        return [k for k in OperationType.values() if k in self.config.keys()], 'default'

    def check_config(self):
        """check the configuration as much as we can as early as we can

        Raises:
            various exceptions if any checks fail

        """
        self.check_metadata()

        self.check_sections()

        self.nodename_to_classname = {}

        unique_class_keys = set()

        self.instance_to_config = {}

        # check that all child nodes of each section have a Factory.CLASS_KEY field
        for section_key in self.config.keys():

            for child_key in self.config[section_key].keys():

                child = self.config[section_key][child_key]

                self.instance_to_config[child_key] = child

                if not NodeFactory.CLASS_KEY in child:
                    raise ConfigurationError("No class key found in %s.%s" % (section_key,child_key))

                self.nodename_to_classname[child_key] = child[NodeFactory.CLASS_KEY]

                unique_class_keys.add((child[NodeFactory.CLASS_KEY], child.get(NodeFactory.CLASS_PREFIX)))

                for k in ['destination_pipeline', 'destination_models', 'destination_postprocesses', 'destination_writer']:
                    if k in child:
                        raise Exception("Do you have a old config file? You have %s. Nodes just have 'destinations':[] now", k)

        logging.info("OK: all class keys are present")

        # get class_prefixes by traversing node package
        unique_class_keys = self._traverse_node_package(unique_class_keys)

        # check that each referenced class is registered in NodeFactory
        for class_key, class_prefix in unique_class_keys:
            if not NodeFactory().is_registered(class_key):
                try:
                    logging.info(f'attempting to register {class_key}')
                    self._register_class(class_key, class_prefix)
                except:
                    raise ConfigurationError(f"Cannot register node class {class_key}")

        #check necessary_configs
        for instance_name in self.nodename_to_classname:
            class_key = self.nodename_to_classname[instance_name]

            configuration_dict = self.instance_to_config[instance_name]

            instance = NodeFactory().instantiate(class_key, self, instance_name)

            NodeFactory().valid_configuration(instance, configuration_dict)

        logging.info("OK: all classes recognized")
        logging.info("OK: good necessary_configs")

        # run our DAG checks. Throws error if not OK
        self.dag.check_dag()

    def _register_class(self, class_key, class_prefix):
        """Register a class specified in the config file.

        Args:
            class_key (str): class key to register
            class_prefix(str): the prefix of the class to register. Can be in `path.to.module` format,
                or a full path `path/to/module`.

        Returns:
            None - attempts to register the class with it's default name
        """
        # convert to string before checking if file
        if class_prefix is None:
            class_prefix = ''

        if os.path.isfile(class_prefix):
            modulename = self._import_file(class_key, class_prefix)

        # loading from module
        else:
            if self.config_metadata and 'class_package' in self.config_metadata:
                class_package = self.config_metadata['class_package']
                prefix = '.'.join(filter(None, [class_package, class_prefix]))
            else:
                prefix = class_prefix

            modulename = importlib.import_module(prefix)

        clz = getattr(modulename, class_key)
        NodeFactory().register(None, clz)

    @staticmethod
    def _import_file(full_name, path):
        """Import module from given path

        Args:
            full_name (str): full module name to import
            path (str): full path to python module

        Returns:
            module imported from the path with given name
        """
        spec = importlib.util.spec_from_file_location(full_name, path)
        mod = importlib.util.module_from_spec(spec)

        spec.loader.exec_module(mod)
        return mod

    def _get_file_candidates(self):
        """Get file candidates to search through when specifying a class package.

        Priority will first consider environment variable PRIMROSE_EXT_NODE_PACKAGE. If unset, will
        search the configuration metadata for key `class_package`. If nothing is specified, in either
        location, an empty list is returned.

        Returns:
            list of potential files to search for classes to register
        """
        # for now assume packages/top level only
        if CLASS_ENV_PACKAGE_KEY in os.environ:
            pkg_name = os.environ[CLASS_ENV_PACKAGE_KEY]
        elif self.config_metadata:
            if 'class_package' in self.config_metadata:
                pkg_name = self.config_metadata['class_package']
            else:
                return []
        else:
            return []
        # look for path to module to find potential file candidates
        try:
            # if we are passed something like __init__.py, grab the package
            if os.path.isfile(pkg_name):
                pkg_name = os.path.dirname(pkg_name)
            # if we have an actual package from pip install
            if not os.path.isdir(pkg_name):
                pkg_name = os.path.dirname(importlib.import_module(pkg_name).__file__)
        except ModuleNotFoundError:
            logging.warning("Could not find module specified for external node configuration")
            return []

        candidates = glob.glob(os.path.join(pkg_name, '**', '*.py'), recursive=True)

        return candidates

    def _traverse_node_package(self, unique_class_keys, overwrite=False):
        """Traverse node package to find classes in the DAG to register.

        Args:
            unique_class_keys (tuple(str, str)): a tuple of class names and prefixes.
            overwrite (boolean, Optional): If a prefix is already set from the configuration, do we overwrite?

        Returns:
            (class_name, class_prefix) tuples
        """
        class_keys_prefix = []
        candidates = self._get_file_candidates()
        for filename in candidates:
            with open(filename, 'r') as f:
                src_str = f.read()
                for class_key, class_key_prefix in unique_class_keys:
                    if (class_key_prefix is None) or (overwrite == True):
                        pattern = "class\s" + class_key + "\(?.*\)?:\s"
                        if re.search(pattern, src_str) is not None:
                            class_keys_prefix.append((class_key, filename))
                            continue
        for class_key, class_key_prefix in unique_class_keys:
            if class_key not in [x[0] for x in class_keys_prefix]:
                class_keys_prefix.append((class_key, class_key_prefix))
        return class_keys_prefix

    def _parse_config(self):
        """Assign top level keys to config attributes

        Note:
            Assign top level keys to config attributes. The method then assigns the inner dict to the top level key object

        Returns:
            Nothing, side effect is that top level keys are added as config attributes

        """
        # __init__ json.load checks that implementation_config top-level keys are unique.
        # Here, we check next-level down and error out on first duplicate found
        all_keys = set()
        for key in self.config.keys():
            section_dict = self.config[key]
            self.__setattr__(key, section_dict)
            for k2 in section_dict.keys():
                if k2 in all_keys:
                    raise ConfigurationError("Operations must all have unique names in the configuration. Duplicate key: '%s'" % k2)
                else:
                    all_keys.add(k2)




import pytest
import pandas as pd

from primrose.transformers.filter import FilterByPandasExpression

def test_transform():
    df = pd.read_csv("test/tennis.csv")

    ft = FilterByPandasExpression([["outlook", "==", "sunny"]])

    assert not ft.fit(df)

    assert df.shape == (14, 6)

    filtered_df = ft.transform(df)

    assert filtered_df.shape == (5, 6)

def test_transform2():
    df = pd.read_csv("test/tennis.csv")

    ft = FilterByPandasExpression([])

    assert df.shape == (14, 6)

    filtered_df = ft.transform(df)

    assert filtered_df.shape == df.shape

def test_transform3():
    ft = FilterByPandasExpression([["outlook", "==", "sunny"]])
    with pytest.raises(Exception) as e:
        ft.transform(42)
    assert "Data is not a pandas DataFrame" in str(e)

def test_transform4():
    df = pd.read_csv("test/tennis.csv")
    ft = FilterByPandasExpression([["JUNK", "==", "sunny"]])
    with pytest.raises(Exception) as e:
        ft.transform(df)
    assert "Unrecognized filter column 'JUNK'" in str(e)

def test_transform5():
    df = pd.read_csv("test/tennis.csv")
    ft = FilterByPandasExpression([["outlook", "JUNK", "sunny"]])
    with pytest.raises(Exception) as e:
        ft.transform(df)
    assert "Unsupported filter operation 'JUNK'" in str(e)





import pytest
import os
from primrose.readers.mysql_helper import MySQLHelper

def test_extract_mongo_credentials():
    keys=["MYSQL_HOST",
            "MYSQL_PORT",
            "MYSQL_DB",
            "MYSQL_USER",
            "MYSQL_PASS"]
    for k in keys:
        if k in os.environ:
            del os.environ[k]

    with pytest.raises(Exception) as e:
        MySQLHelper.extract_mysql_credentials()
    assert 'Did not find env variable for MYSQL_HOST' in str(e)

def test_extract_mongo_credentials2():
    keys=["MYSQL_HOST",
            "MYSQL_PORT",
            "MYSQL_DB",
            "MYSQL_USER",
            "MYSQL_PASS"]
    for i,k in enumerate(keys):
        os.environ[k] = str(i)

    host, port, username, password, database = MySQLHelper.extract_mysql_credentials()
    assert host == "0"
    assert port == 1
    assert username == "3"
    assert password == "4"
    assert database == "2" 


"""Enum to speacify type of run mode: train, predict, and eval

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from enum import Enum

class RunModes(Enum):
    """set of operation type identifiers"""
    TRAIN = 'train'
    PREDICT = 'predict'
    EVAL = 'eval'

    @staticmethod
    def names():
        """list of all the names in the enum

        Returns:
            list of Enum's names

        """
        return list(map(lambda t: t.name, RunModes))

    @staticmethod
    def values():
        """list of all the values in the enum

        Returns:
            list of Enum's values

        """
        return list(map(lambda t: t.value, RunModes))




import pytest
from primrose.base.transformer_sequence import TransformerSequence
from primrose.base.transformer import AbstractTransformer

@pytest.fixture
def transformer_class():
    class TestTransformer(AbstractTransformer):
        def fit(self, data):
            return data
        def transform(self, data):
            return data
    return TestTransformer()

def test_init(transformer_class):
    t1 = transformer_class
    t2 = transformer_class
    ts = TransformerSequence([t1, t2])
    array = list(ts.transformers())
    assert len(array) == 2
    assert array[0] == t1

def test_add(transformer_class):
    ts = TransformerSequence()
    assert len(ts.sequence) == 0
    t = transformer_class
    ts.add(t)
    assert len(ts.sequence) == 1
    
    with pytest.raises(Exception) as e:
        ts.add(0)
    assert 'Transformer needs to extend AbstractTransformer' in str(e)

    ts.add(transformer_class)
    ts.add(transformer_class)

    array = list(ts.transformers())
    assert len(array) == 3
    assert array[0] == t


import pytest
import pandas as pd

from primrose.transformers.strings import StringTransformer

@pytest.fixture
def dataframe():
    df = pd.DataFrame({
        "caps":['TEST!','TEST@'],
        "splits":['a,b,c','def'],
        "strips":['aaa,,,,',',,,bbb'],
        "caps2":['TEST#','TEST$']})
    return df

def test_StringTransformer_single_column_str(dataframe):
    df_test = dataframe.copy()
    df_test['splits'] = df_test['splits'].str.split(',')
    t = StringTransformer('split','splits',',')
    df_out = t.transform(dataframe)
    pd.testing.assert_frame_equal(df_test,df_out)

def test_StringTransformer_multiple_column_list(dataframe):
    df_test = dataframe.copy()
    df_test['caps'] = df_test['caps'].str.lower()
    df_test['caps2'] = df_test['caps2'].str.lower()
    t = StringTransformer('lower',['caps','caps2'])
    df_out = t.transform(dataframe)
    pd.testing.assert_frame_equal(df_test,df_out)

def test_StringTransformer_kwargs(dataframe):
    df_test = dataframe.copy()
    df_test['strips'] = df_test['strips'].str.rstrip(to_strip=',')
    t = StringTransformer('rstrip', ['strips'],to_strip=',')
    df_out = t.transform(dataframe)
    pd.testing.assert_frame_equal(df_test, df_out)

def test__execute_str_method(dataframe):
    df_test = dataframe.copy()
    test_series = df_test['splits'].str.split(',')
    t = StringTransformer('split', 'splits', ',')
    result_series = t._execute_str_method(df_test['splits'])
    pd.testing.assert_series_equal(result_series,test_series)



import pytest
import logging
from primrose.node_factory import NodeFactory
from primrose.base.model import AbstractModel
from primrose.data_object import DataObject
from testfixtures import LogCapture
from primrose.configuration.configuration import Configuration
from primrose.readers.csv_reader import CsvReader
import pandas as pd

def test_run():
    class TestModel(AbstractModel):

        @staticmethod 
        def necessary_config(node_config): return set(['mode']) 

        def train_model(self, data_object):
            logging.info("TRAIN called")
            return data_object

        def eval_model(self, data_object):
            logging.info("EVAL called")
            return data_object

        def predict(self, data_object):
            logging.info("PREDICT called")
            return data_object

    NodeFactory().register("TestModel", TestModel)

    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mymodel"]
                }
            },
            "model_config": {
                "mymodel": {
                    "class": "TestModel",
                    "mode": "train"
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    reader = CsvReader(configuration, "myreader")
    df = pd.read_csv("test/minimal.csv")

    data_object.add(reader, df)

    model = TestModel(configuration, "mymodel")


    with LogCapture() as l:
        model.run(data_object)
    l.check(
        ('root', 'INFO', 'TRAIN called'),
        ('root', 'INFO', 'EVAL called'),
        ('root', 'INFO', 'PREDICT called')
    )


    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mymodel"]
                }
            },
            "model_config": {
                "mymodel": {
                    "class": "TestModel",
                    "mode": "eval"
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    reader = CsvReader(configuration, "myreader")

    data_object.add(reader, df)

    model = TestModel(configuration, "mymodel")

    with LogCapture() as l:
        model.run(data_object)
    l.check(
        ('root', 'INFO', 'EVAL called'),
        ('root', 'INFO', 'PREDICT called')
    )



"""Module toJoin upstream dataframes

Author(s): 
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

"""
import logging
from primrose.base.pipeline import AbstractPipeline, PipelineModeType
from primrose.transformers.combine import LeftJoinDataCombiner
from primrose.base.transformer_sequence import TransformerSequence
from primrose.data_object import DataObject

class DataFrameJoiner(AbstractPipeline):
    """Join upstream dataframes"""

    @staticmethod
    def necessary_config(node_config):
        """Return the necessary configuration keys for the DataFrameJoiner object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            start_table: first table index in alpha order which defines who is eligible for this analysis (all other tables will be left joined to this)
            join_key: list of column names to join dataframes from different readers on

        Returns:
            set of keys

        """
        return set(['start_table', 'join_key'])

    def init_pipeline(self):
        """create the pipeline's TransformerSequence

        Returns:
            a TransformerSequence

        """
        ts = TransformerSequence()
        # Note: this is a trasnformer that does not striclty adhere to Transformer interface
        # It takes a in *list* of data frames, not a single data, and returns a single dataframe
        ts.add(LeftJoinDataCombiner(self.node_config['join_key']))
        return ts

    def transform(self, data_object):
        """Get DataFrames from the data object then put them into a alphabetical list for constant join order

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """

        upstream_data = data_object.get_upstream_data(self.instance_name,
                                                pop_data=False)

        upstream_keys = sorted([k for k in upstream_data.keys()])

        start_key = self.node_config['start_table']

        if start_key not in upstream_keys:
            raise Exception("Could not find start_table in upstream keys: " + start_key)

        upstream_keys = [start_key] + [k for k in upstream_keys if k != start_key]

        logging.info('Joining dataframes from {} sources'.format(len(upstream_keys)))

        # note we are passing in a list of dataframes here
        # note: we assume default DATA_KEY
        data = self.execute_pipeline([upstream_data[k][DataObject.DATA_KEY] for k in upstream_keys], PipelineModeType.TRANSFORM)

        logging.info('Final size of joined data is {} rows.'.format(len(data)))

        data_object.add(self, data, overwrite=False)

        return data_object



"""set of utility methods and enum for configurations

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from enum import Enum

class ConfigurationError(Exception):
    ''' named error specifically for configuration errors'''
    pass

########################################################################
# note: am adding this Enum in file separate from configuration.py 
# to fend off circular dependencies:
#   configuration -> factory -> individual operation classes -> configuration
########################################################################

class ConfigurationSectionType(Enum):
    """set of top-level sections in config"""
    METADATA = 'metadata'
    IMPLEMENTATION_CONFIG = 'implementation_config'

    @staticmethod
    def values():
        '''list of the enum's values'''
        return list(map(lambda t: t.value, ConfigurationSectionType))

class OperationType(Enum):
    """set of operation type identifiers"""
    reader = 'reader_config'
    pipeline = 'pipeline_config'
    model = 'model_config'
    postprocess = 'postprocess_config'
    writer = 'writer_config'
    dataviz = 'dataviz_config'
    cleanup = 'cleanup_config'

    @staticmethod
    def names():
        """list of the enum's names

        Returns:
            list of (name, OperationType) pairs

        """
        return list(map(lambda t: t.name, OperationType))

    @staticmethod
    def values():
        """list of the enum's values

        Returns:
            list of (value, OperationType) pairs

        """
        return list(map(lambda t: t.value, OperationType))

    @staticmethod
    def values_to_names():
        """value:name dictionary

        Returns:
            dict of {value: name}

        """
        return dict(zip(OperationType.values(), OperationType.names()))



"""write some data to a local or gcs file, serialized with dill or pickle.

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)
    Brian Graham (brian.graham@weightwatchers.com)

"""
import dill
import logging
import os
import pickle
from primrose.data_object import DataObjectResponseType
from primrose.writers.abstract_file_writer import AbstractFileWriter


class Serializer(AbstractFileWriter):
    """Serialize some object to a local file."""
    SUPPORTED_SERIALIZERS = {'dill': dill, 'pickle': pickle}

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the Serializer object:

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            serializer (str): 'dill' or 'pickle'

        Returns:
            set of necessary configuration keys

        """

        config_params = set(['serializer'])
        return config_params.union(AbstractFileWriter.necessary_config(node_config))

    def run(self, data_object):
        """serialize some data to a local filesystem using Dill

        Note:
            this will write object under config['key'] in data_object

        Args:
            data_object (DataObject)

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        filename = self.node_config['dir'] + os.path.sep + self.node_config['filename']
        data_key = self.node_config['key']

        logging.info("Saving {} data to {}".format(data_key, filename))

        data_to_write = data_object.get_upstream_data(self.instance_name, pop_data=False, rtype=DataObjectResponseType.KEY_VALUE.value)

        if data_key not in data_to_write:
            raise Exception("Key {} not found inside data_key object".format(data_key))

        if self.node_config['serializer'] not in Serializer.SUPPORTED_SERIALIZERS.keys():
            logging.warning(f"{self.node_config['serializer']} serializer not supported.")
            logging.warning(f"The following serializers are supported {Serializer.SUPPORTED_SERIALIZERS.keys()}.")
            raise Exception(f"Unsupported serializer: {self.node_config['serializer']}")

        serializer = Serializer.SUPPORTED_SERIALIZERS[self.node_config['serializer']]

        with open(filename, 'wb') as f:
            serializer.dump(data_to_write[data_key], f)

        terminate = False

        return data_object, terminate



"""Custom combiner object which works via pandas left merges

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""

import pandas as pd
import logging
from primrose.base.transformer import AbstractTransformer


def left_merge_dataframe_on_validated_join_keys(left_df, right_df, join_keys):
    """Merge two dataframes together or return just the left_df if the right df is None

    Args:
        left_df: valid dataframe with join keys of matching data type (will be validated)
        right_df: None or valid dataframe with join keys of matching data type (will be validated)
        join_keys: list of keys to be validated on datatype and existance in left/right

    Returns:
        joined dataframe object
    """
    if right_df is None:

        initial_data_length = len(left_df)
        logging.info("Initial data length: {}".format(initial_data_length))

        return left_df

    else:

        initial_data_length = len(left_df)

        for j in join_keys:

            if not j in left_df.columns:
                raise Exception('Join key {} not in left {}. Aborting merge.'.format(j, left_df.columns))
            elif not j in right_df.columns:
                raise Exception('Join key {} not in right {}. Aborting merge.'.format(j, right_df.columns))

            if not left_df[j].dtype == right_df[j].dtype:
                logging.info('Join key {} is not of type {}. Casting.'.format(j, right_df[j].dtype))
                try:
                    right_df[j] = right_df[j].astype(left_df[j].dtype)
                except:
                    raise Exception('Cannot cast join key {} as {}'.format(j, left_df[j].dtype))

        left_df = left_df.merge(right_df, on=join_keys, how='left')
        left_df.reset_index(inplace=True, drop=True)

        if len(left_df) > initial_data_length:
            logging.warning("Merge increased data size by {} rows.".format(
                len(left_df) - initial_data_length))

        return left_df


class LeftJoinDataCombiner(AbstractTransformer):
    """combine two dataframes doing a left join"""

    def __init__(self, join_key):
        """initialize the class

        Args:
            join_key (list): the columns to perform the left join on

        """
        self.join_key = join_key

    def fit(self, data):
        """fit the data, here doing nothing
        
        Args:
            data (list)L : list of pandas data frames

        Returns:
            nothing
        """
        pass

    def transform(self, data):
        """Applies left joins to combine dataframes from different readers

        Args:
            data (list): list of dataframes

        Returns:
            dataframe

        """
        if not isinstance(data, list):
            raise Exception("In this transformer, data needs to be a list of dataframes")

        if len(data) < 2:

            logging.warning("Combiner needs at least two reader inputs, passing unchanged data.")

            return data

        elif not isinstance(data[0], pd.DataFrame):

            raise Exception('LeftJoinDataCombiner must operate on an iterable of pandas.DataFrame objects.')

        else:

            combined_data = None

            # loop over each reader key and merge into a single combined dataframe
            for d in data:
                combined_data = d if combined_data is None else left_merge_dataframe_on_validated_join_keys(
                    combined_data, d, self.join_key)

            return combined_data




import sys
import math
import pytest
import pandas as pd
from primrose.configuration.configuration import Configuration
from primrose.base.search_engine import AbstractSearchEngine
from primrose.base.node import AbstractNode
from nltk import ngrams
from primrose.node_factory import NodeFactory
from primrose.data_object import DataObject, DataObjectResponseType

'''we can't instatiate AbstractSearchEngine but this will test some of the methods'''

def test_cosine_similarity_matrix():

    class Testpipeline(AbstractNode):
        def __init__(self, configuration, instance_name):
            self.configuration = configuration
            self.instance_name = instance_name
        @staticmethod
        def necessary_config(node_config):
            return set([])
        def run(self, data_object):
            return data_object, False
    NodeFactory().register("Testpipeline", Testpipeline)

    class TestSimpleSearchEngine(AbstractSearchEngine):
        '''
            simple TFIDF search engine
        '''

        def __init__(self, configuration, instance_name):
            AbstractSearchEngine.__init__(self, configuration, instance_name)

        def tokenize(self, s, stopwords=[], add_ngrams=True):
            q = s.lower()
            tokens = q.replace("-", " ").replace(",", "").replace("(", "").replace(")", "").split(" ")
            tokens = [w for w in tokens if w not in stopwords]
            if add_ngrams:
                bigrams = list(ngrams(tokens, 2))
                strbigrams = ["_".join(t) for t in bigrams]
                tokens.extend(strbigrams)
            return tokens

        def eval_model(data_object): return data_object

    NodeFactory().register("TestSimpleSearchEngine", TestSimpleSearchEngine)

    config = {
          "implementation_config":{
            "pipeline_config": {
                "pipeline1": {
                    "class": "Testpipeline",
                    "destinations": ["recipe_name_model"]
                }
            },
            "model_config": {
                "recipe_name_model": {
                    "class": "TestSimpleSearchEngine",
                    "id_key": "id",
                    "doc_key": "name",
                    "mode": "precict",
                    "destinations": []
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    #set that pipeline provided the corpus
    corpus = [{"id":1,"name":"spinach omelet"},
        {"id":2,"name":"kale omelet"},
        {"id":3,"name":"cherry pie"}]
    data_object = DataObject(configuration)
    pipeline = Testpipeline(configuration, 'pipeline1')
    data_object.add(pipeline, pd.DataFrame(corpus))

    engine = TestSimpleSearchEngine(configuration, 'recipe_name_model')
    engine.predict(data_object)

    m = engine.cosine_similarity_matrix()

    assert math.isclose(m[0, 0], 1., abs_tol=0.001)
    assert math.isclose(m[0, 1], 0.224325, abs_tol=0.001)
    assert math.isclose(m[0, 2], 0., abs_tol=0.001)

    assert math.isclose(m[1, 0], 0.224325, abs_tol=0.001)
    assert math.isclose(m[1, 1], 1., abs_tol=0.001)
    assert math.isclose(m[1, 2], 0., abs_tol=0.001)

    assert math.isclose(m[2, 0], 0., abs_tol=0.001)
    assert math.isclose(m[2, 1], 0., abs_tol=0.001)
    assert math.isclose(m[2, 2], 1., abs_tol=0.001)

    assert engine.ids == [1,2,3]
    assert engine.docs == ["spinach omelet","kale omelet","cherry pie"]
    assert engine.tfidf is not None


"""Module with AbstractReader implementation, able to read from SQLite

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import sqlite3
from primrose.base.sql_reader import AbstractSqlReader

class SQLiteReader(AbstractSqlReader):
    """Runs SQLite queries into pandas dataframes"""

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            After adding this list, validation automatically occurs before instantiation in the pipeline factory.

        Returns:
            set of keys necessary to run implementation

        """
        return set(['filename', 'query_json'])

    def get_connection(self):
        """return connection to SQLite DB

        Returns:
            connection to SQLite DB file

        """
        return sqlite3.connect(self.node_config['filename'])



"""Module with AbstractReader implementation, able to read from AWS Redshift

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.sql_reader import AbstractSqlReader

class OracleReader(AbstractSqlReader):
    """Runs Redshift queries into pandas dataframes"""

    def get_connection(self):
        """return connection to Redshift DB

        Returns:
            connection to Redshift DB

        """
        raise NotImplementedError("Implement connection creation")
